#!/usr/bin/env python
import os
import sys

from django.core.management import execute_from_command_line


def runtests():
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")
    argv = sys.argv[:1] + ["test"] + sys.argv[1:]
    execute_from_command_line(argv)


if __name__ == "__main__":
    runtests()



#!/usr/bin/env python
import os
import sys

from django.core.management import execute_from_command_line


def runshell():
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")
    execute_from_command_line(sys.argv[:1] + ["migrate", "--noinput", "-v", "0"])
    execute_from_command_line(sys.argv[:1] + ["shell"] + sys.argv[1:])


if __name__ == "__main__":
    runshell()



from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from django.test.client import RequestFactory
from django.utils import html

from django_filters.filterset import FilterSet, filterset_factory
from django_filters.views import FilterView

from .models import Book


@override_settings(ROOT_URLCONF="tests.urls")
class GenericViewTestCase(TestCase):
    def setUp(self):
        Book.objects.create(title="Ender's Game", price="1.00", average_rating=3.0)
        Book.objects.create(title="Rainbow Six", price="1.00", average_rating=3.0)
        Book.objects.create(title="Snowcrash", price="1.00", average_rating=3.0)


class GenericClassBasedViewTests(GenericViewTestCase):
    base_url = "/books/"

    def test_view(self):
        response = self.client.get(self.base_url)
        for b in ["Ender's Game", "Rainbow Six", "Snowcrash"]:
            self.assertContains(response, html.escape(b))

    def test_view_filtering_on_title(self):
        response = self.client.get(self.base_url + "?title=Snowcrash")
        for b in ["Ender's Game", "Rainbow Six"]:
            self.assertNotContains(response, html.escape(b))
        self.assertContains(response, "Snowcrash")

    def test_view_with_filterset_not_model(self):
        factory = RequestFactory()
        request = factory.get(self.base_url)
        filterset = filterset_factory(Book)
        view = FilterView.as_view(filterset_class=filterset)
        response = view(request)
        self.assertEqual(response.status_code, 200)
        for b in ["Ender's Game", "Rainbow Six", "Snowcrash"]:
            self.assertContains(response, html.escape(b))

    def test_view_with_model_no_filterset(self):
        factory = RequestFactory()
        request = factory.get(self.base_url)
        view = FilterView.as_view(model=Book)
        response = view(request)
        self.assertEqual(response.status_code, 200)
        for b in ["Ender's Game", "Rainbow Six", "Snowcrash"]:
            self.assertContains(response, html.escape(b))

    def test_view_with_model_and_fields_no_filterset(self):
        factory = RequestFactory()
        request = factory.get(self.base_url + "?price=1.0")
        view = FilterView.as_view(model=Book, filterset_fields=["price"])

        # filtering only by price
        response = view(request)
        self.assertEqual(response.status_code, 200)
        for b in ["Ender's Game", "Rainbow Six", "Snowcrash"]:
            self.assertContains(response, html.escape(b))

        # not filtering by title
        request = factory.get(self.base_url + "?title=Snowcrash")
        response = view(request)
        self.assertEqual(response.status_code, 200)
        for b in ["Ender's Game", "Rainbow Six", "Snowcrash"]:
            self.assertContains(response, html.escape(b))

    def test_view_with_strict_errors(self):
        factory = RequestFactory()
        request = factory.get(self.base_url + "?title=Snowcrash&price=four dollars")
        view = FilterView.as_view(model=Book)
        response = view(request)
        titles = [o.title for o in response.context_data["object_list"]]

        self.assertEqual(response.status_code, 200)
        self.assertEqual(titles, [])

    def test_view_with_non_strict_errors(self):
        factory = RequestFactory()
        request = factory.get(self.base_url + "?title=Snowcrash&price=four dollars")
        view = FilterView.as_view(model=Book, strict=False)
        response = view(request)
        titles = [o.title for o in response.context_data["object_list"]]

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            titles,
            ["Snowcrash"],
        )

    def test_view_without_filterset_or_model(self):
        factory = RequestFactory()
        request = factory.get(self.base_url)
        view = FilterView.as_view()
        with self.assertRaises(ImproperlyConfigured):
            view(request)

    def test_view_with_bad_filterset(self):
        class MyFilterSet(FilterSet):
            pass

        factory = RequestFactory()
        request = factory.get(self.base_url)
        view = FilterView.as_view(filterset_class=MyFilterSet)
        with self.assertRaises(ImproperlyConfigured):
            view(request)

    def test_view_with_unbound_filter_form_returns_initial_queryset(self):
        factory = RequestFactory()
        request = factory.get(self.base_url)

        queryset = Book.objects.filter(title="Snowcrash")
        view = FilterView.as_view(model=Book, queryset=queryset)

        response = view(request)
        titles = [o.title for o in response.context_data["object_list"]]

        self.assertEqual(response.status_code, 200)
        self.assertEqual(titles, ["Snowcrash"])


class GenericFunctionalViewTests(GenericViewTestCase):
    base_url = "/books-legacy/"

    def test_view(self):
        response = self.client.get(self.base_url)
        for b in ["Ender's Game", "Rainbow Six", "Snowcrash"]:
            self.assertContains(response, html.escape(b))
        # extra context
        self.assertEqual(response.context_data["foo"], "bar")
        self.assertEqual(response.context_data["bar"], "foo")

    def test_view_filtering_on_price(self):
        response = self.client.get(self.base_url + "?title=Snowcrash")
        for b in ["Ender's Game", "Rainbow Six"]:
            self.assertNotContains(response, html.escape(b))
        self.assertContains(response, "Snowcrash")



from unittest import mock

from django.db import models


class QuerySet(models.QuerySet):
    def __bool__(self):
        return True


class MockQuerySet:
    """
    Generate a mock that is suitably similar to a QuerySet
    """

    def __new__(self):
        m = mock.Mock(spec_set=QuerySet())
        m.filter.return_value = m
        m.all.return_value = m
        return m



from django.forms import NumberInput, Select, TextInput
from django.test import TestCase

from django_filters.widgets import (
    BaseCSVWidget,
    BooleanWidget,
    CSVWidget,
    LinkWidget,
    LookupChoiceWidget,
    QueryArrayWidget,
    RangeWidget,
    SuffixedMultiWidget,
)


class LookupTypeWidgetTests(TestCase):
    def test_widget_requires_field(self):
        with self.assertRaises(TypeError):
            LookupChoiceWidget()

    def test_widget_render(self):
        widgets = [TextInput(), Select(choices=(("a", "a"), ("b", "b")))]
        w = LookupChoiceWidget(widgets)
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <input name="price" type="text" />
            <select name="price_lookup">
                <option value="a">a</option>
                <option value="b">b</option>
            </select>""",
        )

        self.assertHTMLEqual(
            w.render("price", None),
            """
            <input name="price" type="text" />
            <select name="price_lookup">
                <option value="a">a</option>
                <option value="b">b</option>
            </select>""",
        )

        self.assertHTMLEqual(
            w.render("price", ["2", "a"]),
            """
            <input name="price" type="text" value="2" />
            <select name="price_lookup">
                <option selected="selected" value="a">a</option>
                <option value="b">b</option>
            </select>""",
        )


class LinkWidgetTests(TestCase):
    def test_widget_without_choices(self):
        w = LinkWidget()
        self.assertEqual(len(w.choices), 0)
        self.assertHTMLEqual(w.render("price", ""), """<ul />""")

    def test_widget(self):
        choices = (
            ("test-val1", "test-label1"),
            ("test-val2", "test-label2"),
        )
        w = LinkWidget(choices=choices)
        self.assertEqual(len(w.choices), 2)
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <ul>
                <li><a href="?price=test-val1">test-label1</a></li>
                <li><a href="?price=test-val2">test-label2</a></li>
            </ul>""",
        )

        self.assertHTMLEqual(
            w.render("price", None),
            """
            <ul>
                <li><a href="?price=test-val1">test-label1</a></li>
                <li><a href="?price=test-val2">test-label2</a></li>
            </ul>""",
        )

        self.assertHTMLEqual(
            w.render("price", "test-val1"),
            """
            <ul>
                <li><a class="selected"
                       href="?price=test-val1">test-label1</a></li>
                <li><a href="?price=test-val2">test-label2</a></li>
            </ul>""",
        )

    def test_widget_with_option_groups(self):
        choices = (
            (
                "Audio",
                (
                    ("vinyl", "Vinyl"),
                    ("cd", "CD"),
                ),
            ),
            (
                "Video",
                (
                    ("vhs", "VHS Tape"),
                    ("dvd", "DVD"),
                ),
            ),
            ("unknown", "Unknown"),
        )

        w = LinkWidget(choices=choices)
        self.assertHTMLEqual(
            w.render("media", ""),
            """
            <ul>
                <li><a href="?media=vinyl">Vinyl</a></li>
                <li><a href="?media=cd">CD</a></li>
                <li><a href="?media=vhs">VHS Tape</a></li>
                <li><a href="?media=dvd">DVD</a></li>
                <li><a href="?media=unknown">Unknown</a></li>
            </ul>""",
        )

    def test_widget_with_blank_choice(self):
        choices = (
            ("", "---------"),
            ("test-val1", "test-label1"),
            ("test-val2", "test-label2"),
        )

        w = LinkWidget(choices=choices)
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <ul>
                <li><a class="selected" href="?price=">All</a></li>
                <li><a href="?price=test-val1">test-label1</a></li>
                <li><a href="?price=test-val2">test-label2</a></li>
            </ul>""",
        )

    def test_widget_value_from_datadict(self):
        w = LinkWidget()
        data = {"price": "test-val1"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, "test-val1")


class SuffixedMultiWidgetTests(TestCase):
    def test_assertions(self):
        # number of widgets must match suffixes
        with self.assertRaises(AssertionError):
            SuffixedMultiWidget(widgets=[BooleanWidget])

        # suffixes must be unique
        class W(SuffixedMultiWidget):
            suffixes = ["a", "a"]

        with self.assertRaises(AssertionError):
            W(widgets=[BooleanWidget, BooleanWidget])

        # should succeed
        class W(SuffixedMultiWidget):
            suffixes = ["a", "b"]

        W(widgets=[BooleanWidget, BooleanWidget])

    def test_render(self):
        class W(SuffixedMultiWidget):
            suffixes = ["min", "max"]

        w = W(widgets=[TextInput, TextInput])
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <input name="price_min" type="text" />
            <input name="price_max" type="text" />
        """,
        )

        # blank suffix
        class W(SuffixedMultiWidget):
            suffixes = [None, "lookup"]

        w = W(widgets=[TextInput, TextInput])
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <input name="price" type="text" />
            <input name="price_lookup" type="text" />
        """,
        )

    def test_value_from_datadict(self):
        class W(SuffixedMultiWidget):
            suffixes = ["min", "max"]

        w = W(widgets=[TextInput, TextInput])
        result = w.value_from_datadict(
            {
                "price_min": "1",
                "price_max": "2",
            },
            {},
            "price",
        )
        self.assertEqual(result, ["1", "2"])

        result = w.value_from_datadict({}, {}, "price")
        self.assertEqual(result, [None, None])

        # blank suffix
        class W(SuffixedMultiWidget):
            suffixes = ["", "lookup"]

        w = W(widgets=[TextInput, TextInput])
        result = w.value_from_datadict(
            {
                "price": "1",
                "price_lookup": "lt",
            },
            {},
            "price",
        )
        self.assertEqual(result, ["1", "lt"])

    def test_value_omitted_from_data(self):
        class A(SuffixedMultiWidget):
            suffixes = ["b"]

        a = A(widgets=[BooleanWidget])

        result = a.value_omitted_from_data([], None, "test")

        self.assertIsNotNone(result)

    def test_replace_name(self):
        class A(SuffixedMultiWidget):
            suffixes = ["test"]

        a = A(widgets=[None])

        output = '<div name="test123_0"></div>'
        index = 0
        q = a.replace_name(output, index)
        self.assertEqual(q, '<div name="test123_test"></div>')

    def test_decompress_value_none(self):
        class A(SuffixedMultiWidget):
            suffixes = [""]

        a = A(widgets=[None])
        self.assertEqual(a.decompress(None), [None, None])


class RangeWidgetTests(TestCase):
    def test_widget(self):
        w = RangeWidget()
        self.assertEqual(len(w.widgets), 2)
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <input type="text" name="price_min" />
            -
            <input type="text" name="price_max" />""",
        )

        self.assertHTMLEqual(
            w.render("price", slice(5.99, 9.99)),
            """
            <input type="text" name="price_min" value="5.99" />
            -
            <input type="text" name="price_max" value="9.99" />""",
        )

    def test_widget_attributes(self):
        w = RangeWidget(attrs={"type": "date"})
        self.assertEqual(len(w.widgets), 2)
        self.assertHTMLEqual(
            w.render("date", ""),
            """
            <input type="date" name="date_min" />
            -
            <input type="date" name="date_max" />""",
        )


class BooleanWidgetTests(TestCase):
    def test_widget_render(self):
        w = BooleanWidget()
        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <select name="price">
                <option selected="selected" value="">Unknown</option>
                <option value="true">Yes</option>
                <option value="false">No</option>
            </select>""",
        )

    def test_widget_value_from_datadict(self):
        """ """
        w = BooleanWidget()

        trueActive = {"active": "true"}
        result = w.value_from_datadict(trueActive, {}, "active")
        self.assertEqual(result, True)

        falseActive = {"active": "false"}
        result = w.value_from_datadict(falseActive, {}, "active")
        self.assertEqual(result, False)

        result = w.value_from_datadict({}, {}, "active")
        self.assertEqual(result, None)


class BaseCSVWidgetTests(TestCase):
    def test_widget_render(self):
        class NumberCSVWidget(BaseCSVWidget, NumberInput):
            pass

        w = NumberCSVWidget(attrs={"test": "attr"})
        self.assertHTMLEqual(
            w.render("price", None),
            """
            <input type="number" test="attr" name="price" />""",
        )

        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <input type="number" test="attr" name="price" />""",
        )

        self.assertHTMLEqual(
            w.render("price", []),
            """
            <input type="number" test="attr" name="price" />""",
        )

        self.assertHTMLEqual(
            w.render("price", "1"),
            """
            <input type="number" test="attr" name="price" value="1" />""",
        )

        self.assertHTMLEqual(
            w.render("price", "1,2"),
            """
            <input type="number" test="attr" name="price" value="1,2" />""",
        )

        self.assertHTMLEqual(
            w.render("price", ["1", "2"]),
            """
            <input type="text" name="price" value="1,2" />""",
        )

        self.assertHTMLEqual(
            w.render("price", [1, 2]),
            """
            <input type="text" name="price" value="1,2" />""",
        )

    def test_widget_value_from_datadict(self):
        class NumberCSVWidget(BaseCSVWidget, NumberInput):
            pass

        w = NumberCSVWidget()
        data = {"price": None}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, None)

        data = {"price": "1"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1"])

        data = {"price": "1,2"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1", "2"])

        data = {"price": ["1", "2"]}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1", "2"])

        data = {"price": "1,,2"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1", "", "2"])

        data = {"price": "1,"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1", ""])

        data = {"price": ","}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["", ""])

        data = {"price": ""}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        result = w.value_from_datadict({}, {}, "price")
        self.assertEqual(result, None)

    def test_surrogate_class(self):
        class ClassSurrogate(BaseCSVWidget, NumberInput):
            surrogate = NumberInput

        w = ClassSurrogate()
        self.assertIsInstance(w.surrogate, NumberInput)

    def test_surrogate_instance(self):
        class InstanceSurrogate(BaseCSVWidget, NumberInput):
            surrogate = NumberInput()

        w = InstanceSurrogate()
        self.assertIsInstance(w.surrogate, NumberInput)
        self.assertIsNot(InstanceSurrogate.surrogate, w.surrogate)  # deepcopied


class CSVWidgetTests(TestCase):
    def test_widget_render(self):
        w = CSVWidget(attrs={"test": "attr"})
        self.assertHTMLEqual(
            w.render("price", None),
            """
            <input type="text" test="attr" name="price" />""",
        )

        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <input type="text" test="attr" name="price" />""",
        )

        self.assertHTMLEqual(
            w.render("price", []),
            """
            <input type="text" test="attr" name="price" />""",
        )

        self.assertHTMLEqual(
            w.render("price", "1"),
            """
            <input type="text" test="attr" name="price" value="1" />""",
        )

        self.assertHTMLEqual(
            w.render("price", "1,2"),
            """
            <input type="text" test="attr" name="price" value="1,2" />""",
        )

        self.assertHTMLEqual(
            w.render("price", ["1", "2"]),
            """
            <input type="text" test="attr" name="price" value="1,2" />""",
        )

        self.assertHTMLEqual(
            w.render("price", [1, 2]),
            """
            <input type="text" test="attr" name="price" value="1,2" />""",
        )


class CSVSelectTests(TestCase):
    class CSVSelect(BaseCSVWidget, Select):
        pass

    def test_widget(self):
        w = self.CSVSelect(choices=((1, "a"), (2, "b")))
        self.assertHTMLEqual(
            w.render("price", None),
            """
            <select name="price">
                <option value="1">a</option>
                <option value="2">b</option>
            </select>
            """,
        )

        self.assertHTMLEqual(
            w.render("price", ""),
            """
            <select name="price">
                <option value="1">a</option>
                <option value="2">b</option>
            </select>
            """,
        )

        self.assertHTMLEqual(
            w.render("price", "1"),
            """
            <select name="price">
                <option selected="selected" value="1">a</option>
                <option value="2">b</option>
            </select>
            """,
        )

        self.assertHTMLEqual(
            w.render("price", "1,2"),
            """
            <select name="price">
                <option value="1">a</option>
                <option value="2">b</option>
            </select>
            """,
        )

        self.assertHTMLEqual(
            w.render("price", ["1", "2"]),
            """
            <input type="text" name="price" value="1,2" />""",
        )

        self.assertHTMLEqual(
            w.render("price", [1, 2]),
            """
            <input type="text" name="price" value="1,2" />""",
        )


class QueryArrayWidgetTests(TestCase):
    def test_widget_value_from_datadict(self):
        w = QueryArrayWidget()

        # Values can be provided as csv string: ?foo=bar,baz
        data = {"price": None}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        data = {"price": "1"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1"])

        data = {"price": "1,2"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(sorted(result), ["1", "2"])

        data = {"price": "1,,2"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(sorted(result), ["1", "2"])

        data = {"price": "1,"}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1"])

        data = {"price": ","}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        data = {"price": ""}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        result = w.value_from_datadict({}, {}, "price")
        self.assertEqual(result, [])

        # Values can be provided as query array: ?foo[]=bar&foo[]=baz

        data = {"price[]": None}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        data = {"price[]": ["1"]}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1"])

        data = {"price[]": ["1", "2"]}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(sorted(result), ["1", "2"])

        data = {"price[]": ["1", "", "2"]}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(sorted(result), ["1", "2"])

        data = {"price[]": ["1", ""]}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, ["1"])

        data = {"price[]": ["", ""]}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        data = {"price[]": []}
        result = w.value_from_datadict(data, {}, "price")
        self.assertEqual(result, [])

        result = w.value_from_datadict({}, {}, "price")
        self.assertEqual(result, [])



import datetime as dt
import decimal
import unittest
from datetime import datetime, time, timedelta, tzinfo

import django
import pytz
from django import forms
from django.test import TestCase, override_settings
from django.utils import timezone

from django_filters.fields import (
    BaseCSVField,
    BaseRangeField,
    DateRangeField,
    DateTimeRangeField,
    IsoDateTimeField,
    IsoDateTimeRangeField,
    Lookup,
    LookupChoiceField,
    RangeField,
    TimeRangeField,
)
from django_filters.widgets import BaseCSVWidget, CSVWidget, RangeWidget


def to_d(float_value):
    return decimal.Decimal("%.2f" % float_value)


class LookupTests(TestCase):
    def test_empty_attrs(self):
        with self.assertRaisesMessage(ValueError, ""):
            Lookup(None, None)

        with self.assertRaisesMessage(ValueError, ""):
            Lookup("", "")

    def test_empty_value(self):
        with self.assertRaisesMessage(ValueError, ""):
            Lookup("", "exact")

    def test_empty_lookup_expr(self):
        with self.assertRaisesMessage(ValueError, ""):
            Lookup("Value", "")


class RangeFieldTests(TestCase):
    def test_field(self):
        f = RangeField()
        self.assertEqual(len(f.fields), 2)

    def test_clean(self):
        w = RangeWidget()
        f = RangeField(widget=w, required=False)

        self.assertEqual(f.clean(["12.34", "55"]), slice(to_d(12.34), to_d(55)))
        self.assertIsNone(f.clean([]))


class DateRangeFieldTests(TestCase):
    def test_field(self):
        f = DateRangeField()
        self.assertEqual(len(f.fields), 2)

    @override_settings(USE_TZ=False)
    def test_clean(self):
        w = RangeWidget()
        f = DateRangeField(widget=w, required=False)
        self.assertEqual(
            f.clean(["2015-01-01", "2015-01-10"]),
            slice(
                datetime(2015, 1, 1, 0, 0, 0), datetime(2015, 1, 10, 23, 59, 59, 999999)
            ),
        )
        self.assertIsNone(f.clean([]))


class DateTimeRangeFieldTests(TestCase):
    def test_field(self):
        f = DateTimeRangeField()
        self.assertEqual(len(f.fields), 2)

    @override_settings(USE_TZ=False)
    def test_clean(self):
        w = RangeWidget()
        f = DateTimeRangeField(widget=w)
        self.assertEqual(
            f.clean(["2015-01-01 10:30", "2015-01-10 8:45"]),
            slice(datetime(2015, 1, 1, 10, 30, 0), datetime(2015, 1, 10, 8, 45, 0)),
        )


class IsoDateTimeRangeFieldTests(TestCase):
    def test_field(self):
        f = IsoDateTimeRangeField()
        self.assertEqual(len(f.fields), 2)

    def test_clean(self):
        w = RangeWidget()
        f = IsoDateTimeRangeField(widget=w)
        expected = slice(
            datetime(2015, 1, 1, 9, 30, 1, 123000, tzinfo=dt.timezone.utc),
            datetime(2015, 1, 10, 7, 45, 2, 345000, tzinfo=dt.timezone.utc),
        )
        actual = f.clean(
            ["2015-01-01T10:30:01.123000+01:00", "2015-01-10T08:45:02.345000+01:00"]
        )
        self.assertEqual(expected, actual)


class TimeRangeFieldTests(TestCase):
    def test_field(self):
        f = DateRangeField()
        self.assertEqual(len(f.fields), 2)

    def test_clean(self):
        w = RangeWidget()
        f = TimeRangeField(widget=w)

        self.assertEqual(
            f.clean(["10:15", "12:30"]), slice(time(10, 15, 0), time(12, 30, 0))
        )


class LookupChoiceFieldTests(TestCase):
    def test_field(self):
        inner = forms.DecimalField()
        f = LookupChoiceField(inner, [("gt", "gt"), ("lt", "lt")])
        self.assertEqual(len(f.fields), 2)

    def test_clean(self):
        inner = forms.DecimalField()
        f = LookupChoiceField(inner, [("gt", "gt"), ("lt", "lt")], required=False)
        self.assertEqual(f.clean(["12.34", "lt"]), Lookup(to_d(12.34), "lt"))
        self.assertEqual(f.clean([]), None)

        with self.assertRaisesMessage(forms.ValidationError, "Select a lookup."):
            f.clean(["12.34", ""])

    def test_render_used_html5(self):
        inner = forms.DecimalField()
        f = LookupChoiceField(inner, [("gt", "gt"), ("lt", "lt")], empty_label=None)
        self.assertHTMLEqual(
            f.widget.render("price", ""),
            """
            <input type="number" step="any" name="price" />
            <select name="price_lookup">
                <option value="gt">gt</option>
                <option value="lt">lt</option>
            </select>""",
        )
        self.assertHTMLEqual(
            f.widget.render("price", ["abc", "lt"]),
            """
            <input type="number" step="any" name="price" value="abc" />
            <select name="price_lookup">
                <option value="gt">gt</option>
                <option selected="selected" value="lt">lt</option>
            </select>""",
        )


class IsoDateTimeFieldTests(TestCase):
    reference_str = "2015-07-19T13:34:51.759"
    reference_dt = datetime(2015, 7, 19, 13, 34, 51, 759000)
    field = IsoDateTimeField()

    def parse_input(self, value):
        return self.field.strptime(value, IsoDateTimeField.ISO_8601)

    def test_datetime_string_is_parsed(self):
        d = self.parse_input(self.reference_str)
        self.assertTrue(isinstance(d, datetime))

    def test_datetime_string_with_timezone_is_parsed(self):
        d = self.parse_input(self.reference_str + "+01:00")
        self.assertTrue(isinstance(d, datetime))

    def test_datetime_zulu(self):
        d = self.parse_input(self.reference_str + "Z")
        self.assertTrue(isinstance(d, datetime))

    @unittest.skipUnless(django.VERSION < (5, 0), "pytz support removed in Django 5.0")
    @override_settings(TIME_ZONE="UTC")
    def test_datetime_timezone_awareness(self):
        utc, tokyo = pytz.timezone("UTC"), pytz.timezone("Asia/Tokyo")

        # by default, use the server timezone
        reference = utc.localize(self.reference_dt)
        parsed = self.parse_input(self.reference_str)
        self.assertIsInstance(parsed.tzinfo, tzinfo)
        self.assertEqual(parsed, reference)

        # if set, use the active timezone
        reference = tokyo.localize(self.reference_dt)
        with timezone.override(tokyo):
            parsed = self.parse_input(self.reference_str)
        self.assertIsInstance(parsed.tzinfo, tzinfo)
        self.assertEqual(parsed.tzinfo.zone, tokyo.zone)
        self.assertEqual(parsed, reference)

        # if provided, utc offset should have precedence
        reference = utc.localize(self.reference_dt - timedelta(hours=1))
        parsed = self.parse_input(self.reference_str + "+01:00")
        self.assertIsInstance(parsed.tzinfo, tzinfo)
        self.assertEqual(parsed, reference)

    @override_settings(USE_TZ=False)
    def test_datetime_timezone_naivety(self):
        reference = self.reference_dt.replace()

        parsed = self.parse_input(self.reference_str + "+01:00")
        self.assertIsNone(parsed.tzinfo)
        self.assertEqual(parsed, reference - timedelta(hours=1))

        parsed = self.parse_input(self.reference_str)
        self.assertIsNone(parsed.tzinfo)
        self.assertEqual(parsed, reference)

    def test_datetime_non_iso_format(self):
        f = IsoDateTimeField()
        parsed = f.strptime("19-07-2015T51:34:13.759", "%d-%m-%YT%S:%M:%H.%f")
        self.assertTrue(isinstance(parsed, datetime))
        self.assertEqual(parsed, self.reference_dt)

    def test_datetime_wrong_format(self):
        with self.assertRaises(ValueError):
            self.parse_input("19-07-2015T51:34:13.759")


class BaseCSVFieldTests(TestCase):
    class DecimalCSVField(BaseCSVField, forms.DecimalField):
        pass

    def test_clean(self):
        # Filter class sets required=False by default
        field = self.DecimalCSVField(required=False)

        self.assertEqual(field.clean(None), None)
        self.assertEqual(field.clean(""), [])
        self.assertEqual(field.clean(["1"]), [1])
        self.assertEqual(field.clean(["1", "2"]), [1, 2])
        self.assertEqual(field.clean(["1", "2", "3"]), [1, 2, 3])

    def test_validation_error(self):
        field = self.DecimalCSVField()

        msg = "Enter a number."
        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean(["a", "b", "c"])

    def test_required_error(self):
        field = self.DecimalCSVField(required=True)

        msg = "This field is required."
        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean(None)

        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean([""])

    def test_derived_widget(self):
        with self.assertRaises(AssertionError) as excinfo:
            BaseCSVField(widget=RangeWidget())

        msg = str(excinfo.exception)
        self.assertIn("'BaseCSVField.widget' must be a widget class", msg)
        self.assertIn("RangeWidget", msg)

        widget = CSVWidget(attrs={"class": "class"})
        field = BaseCSVField(widget=widget)
        self.assertIsInstance(field.widget, CSVWidget)
        self.assertEqual(field.widget.attrs, {"class": "class"})

        field = BaseCSVField(widget=CSVWidget)
        self.assertIsInstance(field.widget, CSVWidget)

        field = BaseCSVField(widget=forms.Select)
        self.assertIsInstance(field.widget, forms.Select)
        self.assertIsInstance(field.widget, BaseCSVWidget)


class BaseRangeFieldTests(TestCase):
    class DecimalRangeField(BaseRangeField, forms.DecimalField):
        pass

    def test_clean(self):
        # Filter class sets required=False by default
        field = self.DecimalRangeField(required=False)

        self.assertEqual(field.clean(None), None)
        self.assertEqual(field.clean(""), [])
        self.assertEqual(field.clean([]), [])
        self.assertEqual(field.clean(["1", "2"]), [1, 2])

    def test_validation_error(self):
        field = self.DecimalRangeField()

        msg = "Range query expects two values."
        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean(["1"])

        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean(["1", "2", "3"])

    def test_required_error(self):
        field = self.DecimalRangeField(required=True)

        msg = "This field is required."
        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean(None)

        with self.assertRaisesMessage(forms.ValidationError, msg):
            field.clean([""])



from unittest import skipIf

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings

from django_filters.compat import crispy_forms
from django_filters.rest_framework import FilterSet, filters
from django_filters.widgets import BooleanWidget

from ..models import Article, User


class ArticleFilter(FilterSet):
    class Meta:
        model = Article
        fields = ["author"]


class FilterSetFilterForFieldTests(TestCase):
    def test_isodatetimefilter(self):
        field = Article._meta.get_field("published")
        result = FilterSet.filter_for_field(field, "published")
        self.assertIsInstance(result, filters.IsoDateTimeFilter)
        self.assertEqual(result.field_name, "published")

    def test_booleanfilter_widget(self):
        field = User._meta.get_field("is_active")
        result = FilterSet.filter_for_field(field, "is_active")
        self.assertIsInstance(result, filters.BooleanFilter)
        self.assertEqual(result.extra["widget"], BooleanWidget)

    def test_booleanfilter_widget_nullbooleanfield(self):
        field = User._meta.get_field("is_employed")
        result = FilterSet.filter_for_field(field, "is_employed")
        self.assertIsInstance(result, filters.BooleanFilter)
        self.assertEqual(result.extra["widget"], BooleanWidget)


@skipIf(crispy_forms is None, "django_crispy_forms must be installed")
@override_settings(INSTALLED_APPS=settings.INSTALLED_APPS + ("crispy_forms",))
class CrispyFormsCompatTests(TestCase):
    def test_crispy_helper(self):
        # ensure the helper is present on the form
        self.assertTrue(hasattr(ArticleFilter().form, "helper"))

    def test_form_initialization(self):
        # ensure that crispy compat does not prematurely initialize the form
        self.assertFalse(hasattr(ArticleFilter(), "_form"))



import inspect
from collections import OrderedDict
from datetime import date, datetime, time, timedelta
from unittest import mock

from django import forms
from django.test import TestCase, override_settings
from django.utils import translation
from django.utils.translation import gettext as _

from django_filters import filters, widgets
from django_filters.fields import (
    BaseCSVField,
    DateRangeField,
    DateTimeRangeField,
    IsoDateTimeRangeField,
    Lookup,
    RangeField,
    TimeRangeField,
)
from django_filters.filters import (
    AllValuesFilter,
    BaseCSVFilter,
    BaseInFilter,
    BaseRangeFilter,
    BooleanFilter,
    CharFilter,
    ChoiceFilter,
    DateFilter,
    DateFromToRangeFilter,
    DateRangeFilter,
    DateTimeFilter,
    DateTimeFromToRangeFilter,
    DurationFilter,
    Filter,
    IsoDateTimeFromToRangeFilter,
    LookupChoiceFilter,
    ModelChoiceFilter,
    ModelMultipleChoiceFilter,
    MultipleChoiceFilter,
    NumberFilter,
    NumericRangeFilter,
    OrderingFilter,
    RangeFilter,
    TimeFilter,
    TimeRangeFilter,
    TypedMultipleChoiceFilter,
    UUIDFilter,
)
from tests.models import Book, User


class ModuleImportTests(TestCase):
    def is_filter(self, name, value):
        return isinstance(value, type) and issubclass(value, Filter)

    def test_imports(self):
        # msg = "Expected `filters.%s` to be imported in `filters.__all__`"
        filter_classes = [
            key
            for key, value in inspect.getmembers(filters)
            if isinstance(value, type) and issubclass(value, Filter)
        ]

        # sanity check
        self.assertIn("Filter", filter_classes)
        self.assertIn("BooleanFilter", filter_classes)

        for f in filter_classes:
            self.assertIn(f, filters.__all__)


class FilterTests(TestCase):
    def test_creation(self):
        f = Filter()
        self.assertEqual(f.lookup_expr, "exact")
        self.assertEqual(f.exclude, False)

    def test_creation_order(self):
        f = Filter()
        f2 = Filter()
        self.assertTrue(f2.creation_counter > f.creation_counter)

    def test_default_field(self):
        f = Filter()
        field = f.field
        self.assertIsInstance(field, forms.Field)

    def test_field_with_single_lookup_expr(self):
        f = Filter(lookup_expr="iexact")
        field = f.field
        self.assertIsInstance(field, forms.Field)

    def test_field_params(self):
        with mock.patch.object(Filter, "field_class", spec=["__call__"]) as mocked:
            f = Filter(field_name="somefield", label="somelabel", widget="somewidget")
            f.field
            mocked.assert_called_once_with(
                required=False, label="somelabel", widget="somewidget"
            )

    def test_field_extra_params(self):
        with mock.patch.object(Filter, "field_class", spec=["__call__"]) as mocked:
            f = Filter(someattr="someattr")
            f.field
            mocked.assert_called_once_with(
                required=mock.ANY, label=mock.ANY, someattr="someattr"
            )

    def test_field_required_default(self):
        # filter form fields should not be required by default
        with mock.patch.object(Filter, "field_class", spec=["__call__"]) as mocked:
            f = Filter()
            f.field
            mocked.assert_called_once_with(required=False, label=mock.ANY)

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = Filter()
        result = f.filter(qs, "value")
        qs.filter.assert_called_once_with(None__exact="value")
        self.assertNotEqual(qs, result)

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["filter", "exclude"])
        f = Filter(exclude=True)
        result = f.filter(qs, "value")
        qs.exclude.assert_called_once_with(None__exact="value")
        self.assertNotEqual(qs, result)

    def test_filtering_uses_name(self):
        qs = mock.Mock(spec=["filter"])
        f = Filter(field_name="somefield")
        f.filter(qs, "value")
        result = qs.filter.assert_called_once_with(somefield__exact="value")
        self.assertNotEqual(qs, result)

    def test_filtering_skipped_with_blank_value(self):
        qs = mock.Mock()
        f = Filter()
        result = f.filter(qs, "")
        self.assertListEqual(qs.method_calls, [])
        self.assertEqual(qs, result)

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock()
        f = Filter()
        result = f.filter(qs, None)
        self.assertListEqual(qs.method_calls, [])
        self.assertEqual(qs, result)

    def test_filter_using_method(self):
        qs = mock.NonCallableMock(spec=[])
        method = mock.Mock()
        f = Filter(method=method)
        result = f.filter(qs, "value")
        method.assert_called_once_with(qs, None, "value")
        self.assertNotEqual(qs, result)

    def test_filtering_uses_distinct(self):
        qs = mock.Mock(spec=["filter", "distinct"])
        f = Filter(field_name="somefield", distinct=True)
        f.filter(qs, "value")
        result = qs.distinct.assert_called_once_with()
        self.assertNotEqual(qs, result)


class CharFilterTests(TestCase):
    def test_default_field(self):
        f = CharFilter()
        field = f.field
        self.assertIsInstance(field, forms.CharField)


class UUIDFilterTests(TestCase):
    def test_default_field(self):
        f = UUIDFilter()
        field = f.field
        self.assertIsInstance(field, forms.UUIDField)


class BooleanFilterTests(TestCase):
    def test_default_field(self):
        f = BooleanFilter()
        field = f.field
        self.assertIsInstance(field, forms.NullBooleanField)

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = BooleanFilter(field_name="somefield")
        result = f.filter(qs, True)
        qs.filter.assert_called_once_with(somefield__exact=True)
        self.assertNotEqual(qs, result)

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["exclude"])
        f = BooleanFilter(field_name="somefield", exclude=True)
        result = f.filter(qs, True)
        qs.exclude.assert_called_once_with(somefield__exact=True)
        self.assertNotEqual(qs, result)

    def test_filtering_skipped_with_blank_value(self):
        qs = mock.Mock()
        f = BooleanFilter(field_name="somefield")
        result = f.filter(qs, "")
        self.assertListEqual(qs.method_calls, [])
        self.assertEqual(qs, result)

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock()
        f = BooleanFilter(field_name="somefield")
        result = f.filter(qs, None)
        self.assertListEqual(qs.method_calls, [])
        self.assertEqual(qs, result)

    def test_filtering_lookup_expr(self):
        qs = mock.Mock(spec=["filter"])
        f = BooleanFilter(field_name="somefield", lookup_expr="isnull")
        result = f.filter(qs, True)
        qs.filter.assert_called_once_with(somefield__isnull=True)
        self.assertNotEqual(qs, result)


class ChoiceFilterTests(TestCase):
    def test_default_field(self):
        f = ChoiceFilter()
        field = f.field
        self.assertIsInstance(field, forms.ChoiceField)

    def test_empty_choice(self):
        # default value
        f = ChoiceFilter(choices=[("a", "a")])
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("a", "a"),
            ],
        )

        # set value, allow blank label
        f = ChoiceFilter(choices=[("a", "a")], empty_label="")
        self.assertEqual(
            list(f.field.choices),
            [
                ("", ""),
                ("a", "a"),
            ],
        )

        # disable empty choice w/ None
        f = ChoiceFilter(choices=[("a", "a")], empty_label=None)
        self.assertEqual(
            list(f.field.choices),
            [
                ("a", "a"),
            ],
        )

    def test_null_choice(self):
        # default is to be disabled
        f = ChoiceFilter(
            choices=[("a", "a")],
        )
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("a", "a"),
            ],
        )

        # set label, allow blank label
        f = ChoiceFilter(choices=[("a", "a")], null_label="")
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("null", ""),
                ("a", "a"),
            ],
        )

        # set null value
        f = ChoiceFilter(choices=[("a", "a")], null_value="NULL", null_label="")
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("NULL", ""),
                ("a", "a"),
            ],
        )

        # explicitly disable
        f = ChoiceFilter(choices=[("a", "a")], null_label=None)
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("a", "a"),
            ],
        )

    def test_null_multiplechoice(self):
        # default is to be disabled
        f = MultipleChoiceFilter(
            choices=[("a", "a")],
        )
        self.assertEqual(
            list(f.field.choices),
            [
                ("a", "a"),
            ],
        )

        # set label, allow blank label
        f = MultipleChoiceFilter(choices=[("a", "a")], null_label="")
        self.assertEqual(
            list(f.field.choices),
            [
                ("null", ""),
                ("a", "a"),
            ],
        )

        # set null value
        f = MultipleChoiceFilter(choices=[("a", "a")], null_value="NULL", null_label="")
        self.assertEqual(
            list(f.field.choices),
            [
                ("NULL", ""),
                ("a", "a"),
            ],
        )

        # explicitly disable
        f = MultipleChoiceFilter(choices=[("a", "a")], null_label=None)
        self.assertEqual(
            list(f.field.choices),
            [
                ("a", "a"),
            ],
        )

    @override_settings(
        FILTERS_EMPTY_CHOICE_LABEL="EMPTY LABEL",
        FILTERS_NULL_CHOICE_LABEL="NULL LABEL",
        FILTERS_NULL_CHOICE_VALUE="NULL VALUE",
    )
    def test_settings_overrides(self):
        f = ChoiceFilter(
            choices=[("a", "a")],
        )
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "EMPTY LABEL"),
                ("NULL VALUE", "NULL LABEL"),
                ("a", "a"),
            ],
        )

        f = MultipleChoiceFilter(
            choices=[("a", "a")],
        )
        self.assertEqual(
            list(f.field.choices),
            [
                ("NULL VALUE", "NULL LABEL"),
                ("a", "a"),
            ],
        )

    def test_callable_choices(self):
        def choices():
            yield ("a", "a")
            yield ("b", "b")

        f = ChoiceFilter(choices=choices)
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("a", "a"),
                ("b", "b"),
            ],
        )

    def test_callable_choices_is_lazy(self):
        def choices():
            self.fail("choices should not be called during initialization")

        ChoiceFilter(choices=choices)


class MultipleChoiceFilterTests(TestCase):
    def test_default_field(self):
        f = MultipleChoiceFilter()
        field = f.field
        self.assertIsInstance(field, forms.MultipleChoiceField)

    def test_filtering_requires_name(self):
        qs = mock.Mock(spec=["filter"])
        f = MultipleChoiceFilter()
        with self.assertRaises(TypeError):
            f.filter(qs, ["value"])

    def test_conjoined_default_value(self):
        f = MultipleChoiceFilter()
        self.assertFalse(f.conjoined)

    def test_conjoined_true(self):
        f = MultipleChoiceFilter(conjoined=True)
        self.assertTrue(f.conjoined)

    def test_is_noop_false(self):
        f = MultipleChoiceFilter(required=False)
        f.always_filter = False
        self.assertFalse(f.is_noop(None, ["value"]))

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = MultipleChoiceFilter(field_name="somefield")
        with mock.patch("django_filters.filters.Q") as mockQclass:
            mockQ1, mockQ2 = mock.MagicMock(), mock.MagicMock()
            mockQclass.side_effect = [mockQ1, mockQ2]

            f.filter(qs, ["value"])

            self.assertEqual(
                mockQclass.call_args_list, [mock.call(), mock.call(somefield="value")]
            )
            mockQ1.__ior__.assert_called_once_with(mockQ2)
            qs.filter.assert_called_once_with(mockQ1.__ior__.return_value)
            qs.filter.return_value.distinct.assert_called_once_with()

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["exclude"])
        f = MultipleChoiceFilter(field_name="somefield", exclude=True)
        with mock.patch("django_filters.filters.Q") as mockQclass:
            mockQ1, mockQ2 = mock.MagicMock(), mock.MagicMock()
            mockQclass.side_effect = [mockQ1, mockQ2]

            f.filter(qs, ["value"])

            self.assertEqual(
                mockQclass.call_args_list, [mock.call(), mock.call(somefield="value")]
            )
            mockQ1.__ior__.assert_called_once_with(mockQ2)
            qs.exclude.assert_called_once_with(mockQ1.__ior__.return_value)
            qs.exclude.return_value.distinct.assert_called_once_with()

    def test_filtering_with_lookup_expr(self):
        qs = mock.Mock(spec=["filter"])
        f = MultipleChoiceFilter(field_name="somefield", lookup_expr="icontains")
        with mock.patch("django_filters.filters.Q") as mockQclass:
            mockQ1, mockQ2 = mock.MagicMock(), mock.MagicMock()
            mockQclass.side_effect = [mockQ1, mockQ2]

            f.filter(qs, ["value"])

            self.assertEqual(
                mockQclass.call_args_list,
                [mock.call(), mock.call(somefield__icontains="value")],
            )
            mockQ1.__ior__.assert_called_once_with(mockQ2)
            qs.filter.assert_called_once_with(mockQ1.__ior__.return_value)
            qs.filter.return_value.distinct.assert_called_once_with()

    def test_filtering_on_required_skipped_when_len_of_value_is_len_of_field_choices(
        self,
    ):
        qs = mock.Mock(spec=[])
        f = MultipleChoiceFilter(field_name="somefield", required=True)
        f.always_filter = False
        result = f.filter(qs, [])
        self.assertEqual(len(f.field.choices), 0)
        self.assertEqual(qs, result)

        f.field.choices = ["some", "values", "here"]
        result = f.filter(qs, ["some", "values", "here"])
        self.assertEqual(qs, result)

        result = f.filter(qs, ["other", "values", "there"])
        self.assertEqual(qs, result)

    def test_filtering_skipped_with_empty_list_value_and_some_choices(self):
        qs = mock.Mock(spec=[])
        f = MultipleChoiceFilter(field_name="somefield")
        f.field.choices = ["some", "values", "here"]
        result = f.filter(qs, [])
        self.assertEqual(qs, result)

    def test_filter_conjoined_true(self):
        """Tests that a filter with `conjoined=True` returns objects that
        have all the values included in `value`. For example filter
        users that have all of this books.

        """
        book_kwargs = {"price": 1, "average_rating": 1}
        books = []
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))

        user1 = User.objects.create()
        user2 = User.objects.create()
        user3 = User.objects.create()
        user4 = User.objects.create()
        user5 = User.objects.create()

        user1.favorite_books.add(books[0], books[1])
        user2.favorite_books.add(books[0], books[1], books[2])
        user3.favorite_books.add(books[1], books[2])
        user4.favorite_books.add(books[2], books[3])
        user5.favorite_books.add(books[4], books[5])

        filter_list = (
            (
                (books[0].pk, books[0].pk),  # values
                [1, 2],
            ),  # list of user.pk that have `value` books
            ((books[1].pk, books[1].pk), [1, 2, 3]),
            ((books[2].pk, books[2].pk), [2, 3, 4]),
            (
                (books[3].pk, books[3].pk),
                [
                    4,
                ],
            ),
            (
                (books[4].pk, books[4].pk),
                [
                    5,
                ],
            ),
            ((books[0].pk, books[1].pk), [1, 2]),
            (
                (books[0].pk, books[2].pk),
                [
                    2,
                ],
            ),
            ((books[1].pk, books[2].pk), [2, 3]),
            (
                (books[2].pk, books[3].pk),
                [
                    4,
                ],
            ),
            (
                (books[4].pk, books[5].pk),
                [
                    5,
                ],
            ),
            ((books[3].pk, books[4].pk), []),
        )
        users = User.objects.all()

        for item in filter_list:
            f = MultipleChoiceFilter(field_name="favorite_books__pk", conjoined=True)
            queryset = f.filter(users, item[0])
            expected_pks = [c[0] for c in queryset.values_list("pk")]
            self.assertListEqual(
                expected_pks,
                item[1],
                "Lists Differ: {0} != {1} for case {2}".format(
                    expected_pks, item[1], item[0]
                ),
            )


class TypedMultipleChoiceFilterTests(TestCase):
    def test_default_field(self):
        f = TypedMultipleChoiceFilter()
        field = f.field
        self.assertIsInstance(field, forms.TypedMultipleChoiceField)

    def test_filtering_requires_name(self):
        qs = mock.Mock(spec=["filter"])
        f = TypedMultipleChoiceFilter()
        with self.assertRaises(TypeError):
            f.filter(qs, ["value"])

    def test_conjoined_default_value(self):
        f = TypedMultipleChoiceFilter()
        self.assertFalse(f.conjoined)

    def test_conjoined_true(self):
        f = TypedMultipleChoiceFilter(conjoined=True)
        self.assertTrue(f.conjoined)

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = TypedMultipleChoiceFilter(field_name="somefield")
        with mock.patch("django_filters.filters.Q") as mockQclass:
            mockQ1, mockQ2 = mock.MagicMock(), mock.MagicMock()
            mockQclass.side_effect = [mockQ1, mockQ2]

            f.filter(qs, ["value"])

            self.assertEqual(
                mockQclass.call_args_list, [mock.call(), mock.call(somefield="value")]
            )
            mockQ1.__ior__.assert_called_once_with(mockQ2)
            qs.filter.assert_called_once_with(mockQ1.__ior__.return_value)
            qs.filter.return_value.distinct.assert_called_once_with()

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["exclude"])
        f = TypedMultipleChoiceFilter(field_name="somefield", exclude=True)
        with mock.patch("django_filters.filters.Q") as mockQclass:
            mockQ1, mockQ2 = mock.MagicMock(), mock.MagicMock()
            mockQclass.side_effect = [mockQ1, mockQ2]

            f.filter(qs, ["value"])

            self.assertEqual(
                mockQclass.call_args_list, [mock.call(), mock.call(somefield="value")]
            )
            mockQ1.__ior__.assert_called_once_with(mockQ2)
            qs.exclude.assert_called_once_with(mockQ1.__ior__.return_value)
            qs.exclude.return_value.distinct.assert_called_once_with()

    def test_filtering_on_required_skipped_when_len_of_value_is_len_of_field_choices(
        self,
    ):
        qs = mock.Mock(spec=[])
        f = TypedMultipleChoiceFilter(field_name="somefield", required=True)
        f.always_filter = False
        result = f.filter(qs, [])
        self.assertEqual(len(f.field.choices), 0)
        self.assertEqual(qs, result)

        f.field.choices = ["some", "values", "here"]
        result = f.filter(qs, ["some", "values", "here"])
        self.assertEqual(qs, result)

        result = f.filter(qs, ["other", "values", "there"])
        self.assertEqual(qs, result)

    def test_filtering_skipped_with_empty_list_value_and_some_choices(self):
        qs = mock.Mock(spec=[])
        f = TypedMultipleChoiceFilter(field_name="somefield")
        f.field.choices = ["some", "values", "here"]
        result = f.filter(qs, [])
        self.assertEqual(qs, result)

    def test_filter_conjoined_true(self):
        """Tests that a filter with `conjoined=True` returns objects that
        have all the values included in `value`. For example filter
        users that have all of this books.

        """
        book_kwargs = {"price": 1, "average_rating": 1}
        books = []
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))
        books.append(Book.objects.create(**book_kwargs))

        user1 = User.objects.create()
        user2 = User.objects.create()
        user3 = User.objects.create()
        user4 = User.objects.create()
        user5 = User.objects.create()

        user1.favorite_books.add(books[0], books[1])
        user2.favorite_books.add(books[0], books[1], books[2])
        user3.favorite_books.add(books[1], books[2])
        user4.favorite_books.add(books[2], books[3])
        user5.favorite_books.add(books[4], books[5])

        filter_list = (
            (
                (books[0].pk, books[0].pk),  # values
                [1, 2],
            ),  # list of user.pk that have `value` books
            ((books[1].pk, books[1].pk), [1, 2, 3]),
            ((books[2].pk, books[2].pk), [2, 3, 4]),
            (
                (books[3].pk, books[3].pk),
                [
                    4,
                ],
            ),
            (
                (books[4].pk, books[4].pk),
                [
                    5,
                ],
            ),
            ((books[0].pk, books[1].pk), [1, 2]),
            (
                (books[0].pk, books[2].pk),
                [
                    2,
                ],
            ),
            ((books[1].pk, books[2].pk), [2, 3]),
            (
                (books[2].pk, books[3].pk),
                [
                    4,
                ],
            ),
            (
                (books[4].pk, books[5].pk),
                [
                    5,
                ],
            ),
            ((books[3].pk, books[4].pk), []),
        )
        users = User.objects.all()

        for item in filter_list:
            f = TypedMultipleChoiceFilter(
                field_name="favorite_books__pk", conjoined=True
            )
            queryset = f.filter(users, item[0])
            expected_pks = [c[0] for c in queryset.values_list("pk")]
            self.assertListEqual(
                expected_pks,
                item[1],
                "Lists Differ: {0} != {1} for case {2}".format(
                    expected_pks, item[1], item[0]
                ),
            )


class DateFilterTests(TestCase):
    def test_default_field(self):
        f = DateFilter()
        field = f.field
        self.assertIsInstance(field, forms.DateField)


class DateTimeFilterTests(TestCase):
    def test_default_field(self):
        f = DateTimeFilter()
        field = f.field
        self.assertIsInstance(field, forms.DateTimeField)


class TimeFilterTests(TestCase):
    def test_default_field(self):
        f = TimeFilter()
        field = f.field
        self.assertIsInstance(field, forms.TimeField)


class DurationFilterTests(TestCase):
    def test_default_field(self):
        f = DurationFilter()
        field = f.field
        self.assertIsInstance(field, forms.DurationField)


class MockQuerySetMixin:
    def get_mock_queryset(self):
        """
        Configure mock QuerySet with required methods.
        """
        qs = mock.NonCallableMock(spec=[])
        qs.all = mock.Mock(return_value=qs)
        return qs


class ModelChoiceFilterTests(TestCase, MockQuerySetMixin):
    def test_default_field_without_queryset(self):
        f = ModelChoiceFilter()
        with self.assertRaises(TypeError):
            f.field

    @override_settings(
        FILTERS_EMPTY_CHOICE_LABEL="EMPTY",
        FILTERS_NULL_CHOICE_VALUE="NULL",
    )
    def test_empty_choices(self):
        f = ModelChoiceFilter(
            queryset=User.objects.all(), null_value="null", null_label="NULL"
        )
        self.assertEqual(
            list(f.field.choices),
            [
                ("", "EMPTY"),
                ("null", "NULL"),
            ],
        )

    def test_default_field_with_queryset(self):
        qs = self.get_mock_queryset()
        f = ModelChoiceFilter(queryset=qs)
        field = f.field
        self.assertIsInstance(field, forms.ModelChoiceField)
        self.assertEqual(field.queryset, qs)

    def test_callable_queryset(self):
        request = mock.NonCallableMock(spec=[])
        qs = self.get_mock_queryset()

        qs_callable = mock.Mock(return_value=qs)

        f = ModelChoiceFilter(queryset=qs_callable)
        f.parent = mock.Mock(request=request)
        field = f.field

        qs_callable.assert_called_with(request)
        self.assertEqual(field.queryset, qs)

    def test_get_queryset_override(self):
        request = mock.NonCallableMock(spec=[])
        qs = self.get_mock_queryset()

        class F(ModelChoiceFilter):
            get_queryset = mock.create_autospec(
                ModelChoiceFilter.get_queryset, return_value=qs
            )

        f = F()
        f.parent = mock.Mock(request=request)
        field = f.field

        f.get_queryset.assert_called_with(f, request)
        self.assertEqual(field.queryset, qs)


class ModelMultipleChoiceFilterTests(TestCase, MockQuerySetMixin):
    def test_default_field_without_queryset(self):
        f = ModelMultipleChoiceFilter()
        with self.assertRaises(TypeError):
            f.field

    @override_settings(
        FILTERS_EMPTY_CHOICE_LABEL="EMPTY",
        FILTERS_NULL_CHOICE_VALUE="NULL",
    )
    def test_empty_choices(self):
        f = ModelMultipleChoiceFilter(
            queryset=User.objects.all(), null_value="null", null_label="NULL"
        )
        self.assertEqual(
            list(f.field.choices),
            [
                ("null", "NULL"),
            ],
        )

    def test_default_field_with_queryset(self):
        qs = self.get_mock_queryset()
        f = ModelMultipleChoiceFilter(queryset=qs)
        field = f.field
        self.assertIsInstance(field, forms.ModelMultipleChoiceField)
        self.assertEqual(field.queryset, qs)

    def test_filtering_to_field_name(self):
        qs = User.objects.all()
        f = ModelMultipleChoiceFilter(
            field_name="first_name", to_field_name="first_name", queryset=qs
        )
        user = User.objects.create(first_name="Firstname")

        self.assertEqual(f.get_filter_predicate(user), {"first_name": "Firstname"})
        self.assertEqual(
            f.get_filter_predicate("FilterValue"), {"first_name": "FilterValue"}
        )

        self.assertEqual(list(f.filter(qs, ["Firstname"])), [user])
        self.assertEqual(list(f.filter(qs, [user])), [user])

    def test_callable_queryset(self):
        request = mock.NonCallableMock(spec=[])
        qs = self.get_mock_queryset()

        qs_callable = mock.Mock(return_value=qs)

        f = ModelMultipleChoiceFilter(queryset=qs_callable)
        f.parent = mock.Mock(request=request)
        field = f.field

        qs_callable.assert_called_with(request)
        self.assertEqual(field.queryset, qs)


class NumberFilterTests(TestCase):
    def test_default_field(self):
        f = NumberFilter()
        field = f.field
        self.assertIsInstance(field, forms.DecimalField)

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = NumberFilter()
        f.filter(qs, 1)
        qs.filter.assert_called_once_with(None__exact=1)
        # Also test 0 as it once had a bug
        qs.reset_mock()
        f.filter(qs, 0)
        qs.filter.assert_called_once_with(None__exact=0)

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["exclude"])
        f = NumberFilter(exclude=True)
        f.filter(qs, 1)
        qs.exclude.assert_called_once_with(None__exact=1)
        # Also test 0 as it once had a bug
        qs.reset_mock()
        f.filter(qs, 0)
        qs.exclude.assert_called_once_with(None__exact=0)


class NumericRangeFilterTests(TestCase):
    def test_default_field(self):
        f = NumericRangeFilter()
        field = f.field
        self.assertIsInstance(field, RangeField)

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=20, stop=30)
        f = NumericRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__exact=(20, 30))

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["exclude"])
        value = mock.Mock(start=20, stop=30)
        f = NumericRangeFilter(exclude=True)
        f.filter(qs, value)
        qs.exclude.assert_called_once_with(None__exact=(20, 30))

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = NumericRangeFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_field_with_lookup_expr(self):
        qs = mock.Mock()
        value = mock.Mock(start=20, stop=30)
        f = NumericRangeFilter(lookup_expr=("overlap"))
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__overlap=(20, 30))

    def test_zero_to_zero(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=0, stop=0)
        f = NumericRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__exact=(0, 0))

    def test_filtering_startswith(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=20, stop=None)
        f = NumericRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__startswith=20)

    def test_filtering_endswith(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=None, stop=30)
        f = NumericRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__endswith=30)

    def test_filtering_distinct(self):
        f = NumericRangeFilter(distinct=True)

        # range
        qs = mock.Mock()
        f.filter(qs, mock.Mock(start=20, stop=30))
        qs.distinct.assert_called_once()
        qs.distinct.return_value.filter.assert_called_once_with(None__exact=(20, 30))

        # min
        qs = mock.Mock()
        f.filter(qs, mock.Mock(start=20, stop=None))
        qs.distinct.assert_called_once()
        qs.distinct.return_value.filter.assert_called_once_with(None__startswith=20)

        # max
        qs = mock.Mock()
        f.filter(qs, mock.Mock(start=None, stop=30))
        qs.distinct.assert_called_once()
        qs.distinct.return_value.filter.assert_called_once_with(None__endswith=30)


class RangeFilterTests(TestCase):
    def test_default_field(self):
        f = RangeFilter()
        field = f.field
        self.assertIsInstance(field, RangeField)

    def test_filtering_range(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=20, stop=30)
        f = RangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__range=(20, 30))

    def test_filtering_exclude(self):
        qs = mock.Mock(spec=["exclude"])
        value = mock.Mock(start=20, stop=30)
        f = RangeFilter(exclude=True)
        f.filter(qs, value)
        qs.exclude.assert_called_once_with(None__range=(20, 30))

    def test_filtering_start(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=20, stop=None)
        f = RangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__gte=20)

    def test_filtering_stop(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=None, stop=30)
        f = RangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__lte=30)

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = RangeFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_filtering_ignores_lookup_expr(self):
        qs = mock.Mock()
        value = mock.Mock(start=20, stop=30)
        f = RangeFilter(lookup_expr="gte")
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__range=(20, 30))

    def test_filtering_distinct(self):
        f = RangeFilter(distinct=True)

        # range
        qs = mock.Mock()
        f.filter(qs, mock.Mock(start=20, stop=30))
        qs.distinct.assert_called_once()
        qs.distinct.return_value.filter.assert_called_once_with(None__range=(20, 30))

        # min
        qs = mock.Mock()
        f.filter(qs, mock.Mock(start=20, stop=None))
        qs.distinct.assert_called_once()
        qs.distinct.return_value.filter.assert_called_once_with(None__gte=20)

        # max
        qs = mock.Mock()
        f.filter(qs, mock.Mock(start=None, stop=30))
        qs.distinct.assert_called_once()
        qs.distinct.return_value.filter.assert_called_once_with(None__lte=30)


class DateRangeFilterTests(TestCase):
    def test_creating(self):
        f = DateRangeFilter(empty_label=None)
        self.assertEqual(len(f.choices), 5)
        self.assertIs(f.choices, f.extra["choices"])

        f = DateRangeFilter(empty_label=None, choices=[], filters=[])
        self.assertEqual(f.choices, [])
        self.assertEqual(f.filters, [])
        self.assertEqual(len(f.choices), 0)
        self.assertIs(f.choices, f.extra["choices"])

    def test_default_field(self):
        f = DateRangeFilter()
        field = f.field
        self.assertIsInstance(field, forms.ChoiceField)

    def test_filtering(self):
        # skip filtering, as it's an empty value
        qs = mock.Mock(spec=[])
        f = DateRangeFilter()
        result = f.filter(qs, "")
        self.assertEqual(qs, result)

    def test_filtering_skipped_with_out_of_range_value(self):
        # Field validation should prevent this from occurring
        qs = mock.Mock(spec=[])
        f = DateRangeFilter()
        with self.assertRaises(AssertionError):
            f.filter(qs, "tomorrow")

    def test_choices_and_filters_mismatch(self):
        msg = (
            "Keys must be present in both 'choices' and 'filters'. Missing keys: 'a, b'"
        )
        with self.assertRaisesMessage(AssertionError, msg):
            DateRangeFilter(choices=[("a", "a")], filters={"b": None})

    def test_choices_with_optgroups_dont_mistmatch(self):
        DateRangeFilter(
            choices=[("group", ("a", "a")), ("b", "b")], filters={"a": None, "b": None}
        )

    def test_filtering_for_this_year(self):
        qs = mock.Mock(spec=["filter"])
        with mock.patch("django_filters.filters.now") as mock_now:
            now_dt = mock_now.return_value
            f = DateRangeFilter()
            f.filter(qs, "year")
            qs.filter.assert_called_once_with(None__year=now_dt.year)

    def test_filtering_for_this_month(self):
        qs = mock.Mock(spec=["filter"])
        with mock.patch("django_filters.filters.now") as mock_now:
            now_dt = mock_now.return_value
            f = DateRangeFilter()
            f.filter(qs, "month")
            qs.filter.assert_called_once_with(
                None__year=now_dt.year, None__month=now_dt.month
            )

    def test_filtering_for_7_days(self):
        qs = mock.Mock(spec=["filter"])
        with mock.patch("django_filters.filters.now"), mock.patch(
            "django_filters.filters.timedelta"
        ) as mock_td, mock.patch("django_filters.filters._truncate") as mock_truncate:
            mock_d1, mock_d2 = mock.MagicMock(), mock.MagicMock()
            mock_truncate.side_effect = [mock_d1, mock_d2]
            f = DateRangeFilter()
            f.filter(qs, "week")
            self.assertEqual(
                mock_td.call_args_list, [mock.call(days=7), mock.call(days=1)]
            )
            qs.filter.assert_called_once_with(None__lt=mock_d2, None__gte=mock_d1)

    def test_filtering_for_today(self):
        qs = mock.Mock(spec=["filter"])
        with mock.patch("django_filters.filters.now") as mock_now:
            now_dt = mock_now.return_value
            f = DateRangeFilter()
            f.filter(qs, "today")
            qs.filter.assert_called_once_with(
                None__year=now_dt.year, None__month=now_dt.month, None__day=now_dt.day
            )

    def test_filtering_for_yesterday(self):
        qs = mock.Mock(spec=["filter"])
        with mock.patch("django_filters.filters.now") as mock_now:
            now_dt = mock_now.return_value
            f = DateRangeFilter()
            f.filter(qs, "yesterday")
            qs.filter.assert_called_once_with(
                None__year=(now_dt - timedelta(days=1)).year,
                None__month=(now_dt - timedelta(days=1)).month,
                None__day=(now_dt - timedelta(days=1)).day,
            )


class DateFromToRangeFilterTests(TestCase):
    def test_default_field(self):
        f = DateFromToRangeFilter()
        field = f.field
        self.assertIsInstance(field, DateRangeField)

    def test_filtering_range(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=date(2015, 4, 7), stop=date(2015, 9, 6))
        f = DateFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(
            None__range=(date(2015, 4, 7), date(2015, 9, 6))
        )

    def test_filtering_start(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=date(2015, 4, 7), stop=None)
        f = DateFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__gte=date(2015, 4, 7))

    def test_filtering_stop(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=None, stop=date(2015, 9, 6))
        f = DateFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__lte=date(2015, 9, 6))

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = DateFromToRangeFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_filtering_ignores_lookup_expr(self):
        qs = mock.Mock()
        value = mock.Mock(start=date(2015, 4, 7), stop=date(2015, 9, 6))
        f = DateFromToRangeFilter(lookup_expr="gte")
        f.filter(qs, value)
        qs.filter.assert_called_once_with(
            None__range=(date(2015, 4, 7), date(2015, 9, 6))
        )


class DateTimeFromToRangeFilterTests(TestCase):
    def test_default_field(self):
        f = DateTimeFromToRangeFilter()
        field = f.field
        self.assertIsInstance(field, DateTimeRangeField)

    def test_filtering_range(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(
            start=datetime(2015, 4, 7, 8, 30), stop=datetime(2015, 9, 6, 11, 45)
        )
        f = DateTimeFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(
            None__range=(datetime(2015, 4, 7, 8, 30), datetime(2015, 9, 6, 11, 45))
        )

    def test_filtering_start(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=datetime(2015, 4, 7, 8, 30), stop=None)
        f = DateTimeFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__gte=datetime(2015, 4, 7, 8, 30))

    def test_filtering_stop(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=None, stop=datetime(2015, 9, 6, 11, 45))
        f = DateTimeFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__lte=datetime(2015, 9, 6, 11, 45))

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = DateTimeFromToRangeFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_filtering_ignores_lookup_expr(self):
        qs = mock.Mock()
        value = mock.Mock(
            start=datetime(2015, 4, 7, 8, 30), stop=datetime(2015, 9, 6, 11, 45)
        )
        f = DateTimeFromToRangeFilter(lookup_expr="gte")
        f.filter(qs, value)
        qs.filter.assert_called_once_with(
            None__range=(datetime(2015, 4, 7, 8, 30), datetime(2015, 9, 6, 11, 45))
        )


class IsoDateTimeFromToRangeFilterTests(TestCase):
    def test_default_field(self):
        f = IsoDateTimeFromToRangeFilter()
        field = f.field
        self.assertIsInstance(field, IsoDateTimeRangeField)

    def test_filtering_range(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(
            start=datetime(2015, 4, 7, 8, 30), stop=datetime(2015, 9, 6, 11, 45)
        )
        f = IsoDateTimeFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(
            None__range=(datetime(2015, 4, 7, 8, 30), datetime(2015, 9, 6, 11, 45))
        )

    def test_filtering_start(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=datetime(2015, 4, 7, 8, 30), stop=None)
        f = IsoDateTimeFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__gte=datetime(2015, 4, 7, 8, 30))

    def test_filtering_stop(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=None, stop=datetime(2015, 9, 6, 11, 45))
        f = IsoDateTimeFromToRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__lte=datetime(2015, 9, 6, 11, 45))

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = IsoDateTimeFromToRangeFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_filtering_ignores_lookup_expr(self):
        qs = mock.Mock()
        value = mock.Mock(
            start=datetime(2015, 4, 7, 8, 30), stop=datetime(2015, 9, 6, 11, 45)
        )
        f = IsoDateTimeFromToRangeFilter(lookup_expr="gte")
        f.filter(qs, value)
        qs.filter.assert_called_once_with(
            None__range=(datetime(2015, 4, 7, 8, 30), datetime(2015, 9, 6, 11, 45))
        )


class TimeRangeFilterTests(TestCase):
    def test_default_field(self):
        f = TimeRangeFilter()
        field = f.field
        self.assertIsInstance(field, TimeRangeField)

    def test_filtering_range(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=time(10, 15), stop=time(12, 30))
        f = TimeRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__range=(time(10, 15), time(12, 30)))

    def test_filtering_start(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=time(10, 15), stop=None)
        f = TimeRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__gte=time(10, 15))

    def test_filtering_stop(self):
        qs = mock.Mock(spec=["filter"])
        value = mock.Mock(start=None, stop=time(12, 30))
        f = TimeRangeFilter()
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__lte=time(12, 30))

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = TimeRangeFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_filtering_ignores_lookup_expr(self):
        qs = mock.Mock()
        value = mock.Mock(start=time(10, 15), stop=time(12, 30))
        f = TimeRangeFilter(lookup_expr="gte")
        f.filter(qs, value)
        qs.filter.assert_called_once_with(None__range=(time(10, 15), time(12, 30)))


class AllValuesFilterTests(TestCase):
    def test_default_field_without_assigning_model(self):
        f = AllValuesFilter()
        with self.assertRaises(AttributeError):
            f.field

    def test_default_field_with_assigning_model(self):
        mocked = mock.Mock()
        chained_call = ".".join(
            [
                "_default_manager",
                "distinct.return_value",
                "order_by.return_value",
                "values_list.return_value",
            ]
        )
        mocked.configure_mock(**{chained_call: iter([])})
        f = AllValuesFilter()
        f.model = mocked
        field = f.field
        self.assertIsInstance(field, forms.ChoiceField)

    def test_empty_value_in_choices(self):
        f = AllValuesFilter(field_name="username")
        f.model = User

        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
            ],
        )


class LookupChoiceFilterTests(TestCase):
    def test_normalize_lookup_no_display_label(self):
        # display label has underscores replaced and is capitalized
        display_label = LookupChoiceFilter.normalize_lookup("has_key")
        self.assertEqual(display_label, ("has_key", "Has key"))

    def test_normalize_lookup_with_display_label(self):
        # display label is not transformed if provided
        display_label = LookupChoiceFilter.normalize_lookup(("equal", "equals"))
        self.assertEqual(display_label, ("equal", "equals"))

    def test_lookup_choices_default(self):
        # Lookup choices should default to the model field's registered lookups
        f = LookupChoiceFilter(field_name="username", lookup_choices=None)
        f.model = User

        choice_field = f.field.fields[1]
        self.assertEqual(
            len(choice_field.choices),
            len(User._meta.get_field("username").get_lookups()) + 1,
        )

        field_choices = dict(choice_field.choices)
        self.assertEqual(field_choices["exact"], "Exact")
        self.assertEqual(field_choices["startswith"], "Startswith")

    def test_lookup_choices_list(self):
        f = LookupChoiceFilter(
            field_name="username", lookup_choices=["exact", "startswith", "has_key"]
        )

        choice_field = f.field.fields[1]
        self.assertEqual(
            list(choice_field.choices),
            [
                ("", "---------"),
                ("exact", "Exact"),
                ("startswith", "Startswith"),
                ("has_key", "Has key"),
            ],
        )

    def test_lookup_choices_pairs(self):
        f = LookupChoiceFilter(
            field_name="username",
            lookup_choices=[
                ("exact", "Is equal to"),
                ("startswith", "Starts with"),
            ],
        )

        choice_field = f.field.fields[1]
        self.assertEqual(
            list(choice_field.choices),
            [
                ("", "---------"),
                ("exact", "Is equal to"),
                ("startswith", "Starts with"),
            ],
        )

    def test_lookup_choices_empty_label_default(self):
        f = LookupChoiceFilter(field_name="username", lookup_choices=[])

        choice_field = f.field.fields[1]
        self.assertEqual(list(choice_field.choices), [("", "---------")])

    def test_lookup_choices_empty_label_disabled(self):
        f = LookupChoiceFilter(
            field_name="username", empty_label=None, lookup_choices=[]
        )

        choice_field = f.field.fields[1]
        self.assertEqual(list(choice_field.choices), [])

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = LookupChoiceFilter(
            field_name="somefield", lookup_choices=["some_lookup_expr"]
        )
        result = f.filter(qs, Lookup("value", "some_lookup_expr"))
        qs.filter.assert_called_once_with(somefield__some_lookup_expr="value")
        self.assertNotEqual(qs, result)


class CSVFilterTests(TestCase):
    def setUp(self):
        class NumberInFilter(BaseCSVFilter, NumberFilter):
            pass

        class DateTimeYearInFilter(BaseCSVFilter, DateTimeFilter):
            pass

        self.number_in = NumberInFilter(lookup_expr="in")
        self.datetimeyear_in = DateTimeYearInFilter(lookup_expr="year__in")

    def test_default_field(self):
        f = BaseCSVFilter()
        field = f.field
        self.assertIsInstance(field, forms.Field)

    def test_concrete_field(self):
        field = self.number_in.field
        self.assertIsInstance(field, forms.DecimalField)
        self.assertIsInstance(field, BaseCSVField)
        self.assertEqual(field.__class__.__name__, "DecimalInField")

        field = self.datetimeyear_in.field
        self.assertIsInstance(field, forms.DateTimeField)
        self.assertIsInstance(field, BaseCSVField)
        self.assertEqual(field.__class__.__name__, "DateTimeYearInField")

    def test_filtering(self):
        qs = mock.Mock(spec=["filter"])
        f = self.number_in
        f.filter(qs, [1, 2])
        qs.filter.assert_called_once_with(None__in=[1, 2])

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["filter"])
        f = self.number_in
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_field_with_lookup_expr(self):
        qs = mock.Mock()
        f = self.datetimeyear_in
        f.filter(qs, [1, 2])
        qs.filter.assert_called_once_with(None__year__in=[1, 2])


class BaseInFilterTests(TestCase):
    def test_filtering(self):
        class NumberInFilter(BaseInFilter, NumberFilter):
            pass

        qs = mock.Mock(spec=["filter"])
        f = NumberInFilter()
        f.filter(qs, [1, 2])
        qs.filter.assert_called_once_with(None__in=[1, 2])


class BaseRangeFilterTests(TestCase):
    def test_filtering(self):
        class NumberInFilter(BaseRangeFilter, NumberFilter):
            pass

        qs = mock.Mock(spec=["filter"])
        f = NumberInFilter()
        f.filter(qs, [1, 2])
        qs.filter.assert_called_once_with(None__range=[1, 2])


class OrderingFilterTests(TestCase):
    def test_default_field(self):
        f = OrderingFilter()
        field = f.field
        self.assertIsInstance(field, forms.ChoiceField)

    def test_filtering(self):
        qs = mock.Mock(spec=["order_by"])
        f = OrderingFilter()
        f.filter(qs, ["a", "b"])
        qs.order_by.assert_called_once_with("a", "b")

    def test_filtering_descending(self):
        qs = mock.Mock(spec=["order_by"])
        f = OrderingFilter()
        f.filter(qs, ["-a"])
        qs.order_by.assert_called_once_with("-a")

    def test_filtering_with_fields(self):
        qs = mock.Mock(spec=["order_by"])
        f = OrderingFilter(fields={"a": "b"})
        f.filter(qs, ["b", "-b"])
        qs.order_by.assert_called_once_with("a", "-a")

    def test_filtering_skipped_with_none_value(self):
        qs = mock.Mock(spec=["order_by"])
        f = OrderingFilter()
        result = f.filter(qs, None)
        self.assertEqual(qs, result)

    def test_choices_unaltered(self):
        # provided 'choices' should not be altered when 'fields' is present
        f = OrderingFilter(
            choices=(("a", "A"), ("b", "B")),
            fields=(("a", "c"), ("b", "d")),
        )

        self.assertSequenceEqual(
            list(f.field.choices),
            (
                ("", "---------"),
                ("a", "A"),
                ("b", "B"),
            ),
        )

    def test_choices_from_fields(self):
        f = OrderingFilter(
            fields=(("a", "c"), ("b", "d")),
        )

        self.assertSequenceEqual(
            list(f.field.choices),
            (
                ("", "---------"),
                ("c", "C"),
                ("-c", "C (descending)"),
                ("d", "D"),
                ("-d", "D (descending)"),
            ),
        )

    def test_field_labels(self):
        f = OrderingFilter(
            fields=(("a", "c"), ("b", "d")),
            field_labels={"a": "foo"},
        )

        self.assertSequenceEqual(
            list(f.field.choices),
            (
                ("", "---------"),
                ("c", "foo"),
                ("-c", "foo (descending)"),
                ("d", "D"),
                ("-d", "D (descending)"),
            ),
        )

    def test_field_labels_descending(self):
        f = OrderingFilter(
            fields=["username"],
            field_labels={
                "username": "BLABLA",
                "-username": "XYZXYZ",
            },
        )

        self.assertEqual(
            list(f.field.choices),
            [
                ("", "---------"),
                ("username", "BLABLA"),
                ("-username", "XYZXYZ"),
            ],
        )

    def test_normalize_fields(self):
        f = OrderingFilter.normalize_fields
        O = OrderedDict  # noqa

        self.assertIn("a", f({"a": "b"}))

        self.assertEqual(f(O([("a", "b"), ("c", "d")])), O([("a", "b"), ("c", "d")]))

        self.assertEqual(f([("a", "b"), ("c", "d")]), O([("a", "b"), ("c", "d")]))

        self.assertEqual(f(["a", "b"]), O([("a", "a"), ("b", "b")]))

        with self.assertRaises(AssertionError) as ctx:
            f(None)
        self.assertEqual(
            str(ctx.exception),
            "'fields' must be an iterable (e.g., a list, tuple, or mapping).",
        )

        with self.assertRaises(AssertionError) as ctx:
            f([("a", "b", "c")])
        self.assertEqual(
            str(ctx.exception),
            "'fields' must contain strings or (field name, param name) pairs.",
        )

        with self.assertRaises(AssertionError) as ctx:
            f([0, 1, 2])
        self.assertEqual(
            str(ctx.exception),
            "'fields' must contain strings or (field name, param name) pairs.",
        )

    def test_widget(self):
        f = OrderingFilter()
        widget = f.field.widget

        self.assertIsInstance(widget, widgets.BaseCSVWidget)
        self.assertIsInstance(widget, forms.Select)

    def test_translation_sanity(self):
        with translation.override("pl"):
            self.assertEqual(_("Username"), "Nazwa użytkownika")
            self.assertEqual(
                _("%s (descending)") % _("Username"), "Nazwa użytkownika (malejąco)"
            )

    def test_translation_default_label(self):
        with translation.override("pl"):
            f = OrderingFilter(fields=["username"])

            self.assertEqual(
                list(f.field.choices),
                [
                    ("", "---------"),
                    ("username", "Nazwa użytkownika"),
                    ("-username", "Nazwa użytkownika (malejąco)"),
                ],
            )

    def test_translation_override_label(self):
        with translation.override("pl"):
            f = OrderingFilter(
                fields=["username"],
                field_labels={"username": "BLABLA"},
            )

            self.assertEqual(
                list(f.field.choices),
                [
                    ("", "---------"),
                    ("username", "BLABLA"),
                    ("-username", "BLABLA (malejąco)"),
                ],
            )

    def test_help_text(self):
        # regression test for #756 - the usual CSV help_text is not relevant to ordering filters.
        self.assertEqual(OrderingFilter().field.help_text, "")
        self.assertEqual(OrderingFilter(help_text="a").field.help_text, "a")



import datetime
import unittest
import warnings

import django
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.db.models.fields.related import ForeignObjectRel
from django.test import TestCase, override_settings
from django.utils.functional import Promise
from django.utils.timezone import get_default_timezone, make_aware

from django_filters import FilterSet
from django_filters.exceptions import FieldLookupError
from django_filters.filters import MultipleChoiceFilter
from django_filters.utils import (
    MigrationNotice,
    RenameAttributesBase,
    get_field_parts,
    get_model_field,
    handle_timezone,
    label_for_filter,
    resolve_field,
    translate_validation,
    verbose_field_name,
    verbose_lookup_expr,
)

from .models import Article, Book, Business, Company, HiredWorker, NetworkSetting, User


class MigrationNoticeTests(TestCase):
    def test_message(self):
        self.assertEqual(
            str(MigrationNotice("Message.")),
            "Message. See: https://django-filter.readthedocs.io/en/main/guide/migration.html",
        )


class RenameAttributes(RenameAttributesBase):
    renamed_attributes = (("old", "new", DeprecationWarning),)


class SENTINEL:
    pass


class RenameAttributesBaseTests(TestCase):
    def check(self, recorded, count):
        expected = "`Example.old` attribute should be renamed `new`."

        self.assertEqual(len(recorded), count)
        for _ in range(count):
            message = str(recorded.pop().message)
            self.assertEqual(message, expected)
        self.assertEqual(len(recorded), 0)

    def test_class_creation_warnings(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("always")

            class Example(metaclass=RenameAttributes):
                old = SENTINEL

            # single warning for renamed attr on creation
            self.check(recorded, 1)

    def test_renamed_attribute_in_class_dict(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("ignore")

            class Example(metaclass=RenameAttributes):
                old = SENTINEL

            warnings.simplefilter("always")

            # Ensure `old` and `new` are not both in class dict.
            self.assertNotIn("old", Example.__dict__)
            self.assertIn("new", Example.__dict__)

            # Ensure `old` value assigned to `new`.
            self.assertEqual(Example.new, SENTINEL)

            self.check(recorded, 0)

    def test_class_accessor_warnings(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("ignore")

            class Example(metaclass=RenameAttributes):
                new = None

            warnings.simplefilter("always")

            self.assertIsNone(Example.new)
            self.assertIsNone(Example.old)
            self.check(recorded, 1)

            Example.old = SENTINEL
            self.assertIs(Example.new, SENTINEL)
            self.assertIs(Example.old, SENTINEL)
            self.check(recorded, 2)

    def test_instance_accessor_warnings(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("ignore")

            class Example(metaclass=RenameAttributes):
                new = None

            warnings.simplefilter("always")

            example = Example()
            self.check(recorded, 0)

            self.assertIsNone(example.new)
            self.assertIsNone(example.old)
            self.check(recorded, 1)

            example.old = SENTINEL
            self.assertIs(example.new, SENTINEL)
            self.assertIs(example.old, SENTINEL)
            self.check(recorded, 2)

    def test_class_instance_values(self):
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("ignore")

            class Example(metaclass=RenameAttributes):
                new = None

            example = Example()

            # setting instance should not affect class
            example.old = SENTINEL
            self.assertIsNone(Example.old)
            self.assertIsNone(Example.new)
            self.assertIs(example.old, SENTINEL)
            self.assertIs(example.new, SENTINEL)

    def test_getter_reachable(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("always")

            class Example(metaclass=RenameAttributes):
                def __getattr__(self, name):
                    if name == "test":
                        return SENTINEL
                    return self.__getattribute__(name)

            example = Example()
            self.assertIs(example.test, SENTINEL)
            self.check(recorded, 0)

    def test_parent_getter_reachable(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("always")

            class Parent:
                def __getattr__(self, name):
                    if name == "test":
                        return SENTINEL
                    return self.__getattribute__(name)

            class Example(Parent, metaclass=RenameAttributes):
                pass

            example = Example()
            self.assertIs(example.test, SENTINEL)
            self.check(recorded, 0)

    def test_setter_reachable(self):
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter("always")

            class Example(metaclass=RenameAttributes):
                def __setattr__(self, name, value):
                    if name == "test":
                        value = SENTINEL
                    super().__setattr__(name, value)

            example = Example()
            example.test = None
            self.assertIs(example.test, SENTINEL)
            self.check(recorded, 0)


class GetFieldPartsTests(TestCase):
    def test_field(self):
        parts = get_field_parts(User, "username")

        self.assertEqual(len(parts), 1)
        self.assertIsInstance(parts[0], models.CharField)

    def test_non_existent_field(self):
        result = get_model_field(User, "unknown__name")
        self.assertIsNone(result)

    def test_forwards_related_field(self):
        parts = get_field_parts(User, "favorite_books__title")

        self.assertEqual(len(parts), 2)
        self.assertIsInstance(parts[0], models.ManyToManyField)
        self.assertIsInstance(parts[1], models.CharField)

    def test_reverse_related_field(self):
        parts = get_field_parts(User, "manager_of__users__username")

        self.assertEqual(len(parts), 3)
        self.assertIsInstance(parts[0], ForeignObjectRel)
        self.assertIsInstance(parts[1], models.ManyToManyField)
        self.assertIsInstance(parts[2], models.CharField)

    def test_lazy_relationship_not_ready(self):
        """
        This simulates trying to create a FilterSet before the app registry has
        been populated. Lazy relationships have not yet been resolved from their
        strings into their remote model references.
        """

        class TestModel(models.Model):
            fk = models.ForeignKey("remote.Model", on_delete=models.CASCADE)

        msg = (
            "Unable to resolve relationship `fk__f` for `tests.TestModel`. "
            "Django is most likely not initialized, and its apps registry "
            "not populated. Ensure Django has finished setup before loading "
            "`FilterSet`s."
        )
        with self.assertRaisesMessage(RuntimeError, msg):
            get_field_parts(TestModel, "fk__f")


class GetModelFieldTests(TestCase):
    def test_non_existent_field(self):
        result = get_model_field(User, "unknown__name")
        self.assertIsNone(result)

    def test_related_field(self):
        result = get_model_field(Business, "hiredworker__worker")
        self.assertEqual(result, HiredWorker._meta.get_field("worker"))


class ResolveFieldTests(TestCase):
    def test_resolve_plain_lookups(self):
        """
        Check that the standard query terms can be correctly resolved.
        eg, an 'EXACT' lookup on a user's username
        """
        model_field = User._meta.get_field("username")
        lookups = model_field.class_lookups.keys()

        # This is simple - the final output of an untransformed field is itself.
        # The lookups are the default lookups registered to the class.
        for term in lookups:
            field, lookup = resolve_field(model_field, term)
            self.assertIsInstance(field, models.CharField)
            self.assertEqual(lookup, term)

    def test_resolve_forward_related_lookups(self):
        """
        Check that lookups can be resolved for related fields
        in the forwards direction.
        """
        lookups = [
            "exact",
            "gte",
            "gt",
            "lte",
            "lt",
            "in",
            "isnull",
        ]

        # ForeignKey
        model_field = Article._meta.get_field("author")
        for term in lookups:
            field, lookup = resolve_field(model_field, term)
            self.assertIsInstance(field, models.ForeignKey)
            self.assertEqual(lookup, term)

        # ManyToManyField
        model_field = User._meta.get_field("favorite_books")
        for term in lookups:
            field, lookup = resolve_field(model_field, term)
            self.assertIsInstance(field, models.ManyToManyField)
            self.assertEqual(lookup, term)

    def test_resolve_reverse_related_lookups(self):
        """
        Check that lookups can be resolved for related fields
        in the reverse direction.
        """
        lookups = [
            "exact",
            "gte",
            "gt",
            "lte",
            "lt",
            "in",
            "isnull",
        ]

        # ManyToOneRel
        model_field = User._meta.get_field("article")
        for term in lookups:
            field, lookup = resolve_field(model_field, term)
            self.assertIsInstance(field, models.ManyToOneRel)
            self.assertEqual(lookup, term)

        # ManyToManyRel
        model_field = Book._meta.get_field("lovers")
        for term in lookups:
            field, lookup = resolve_field(model_field, term)
            self.assertIsInstance(field, models.ManyToManyRel)
            self.assertEqual(lookup, term)

    def test_resolve_transformed_lookups(self):
        """
        Check that chained field transforms are correctly resolved.
        eg, a 'date__year__gte' lookup on an article's 'published' timestamp.
        """
        # Use a DateTimeField, so we can check multiple transforms.
        # eg, date__year__gte
        model_field = Article._meta.get_field("published")

        standard_lookups = [
            "exact",
            "iexact",
            "gte",
            "gt",
            "lte",
            "lt",
        ]

        date_lookups = [
            "year",
            "month",
            "day",
            "week_day",
        ]

        datetime_lookups = date_lookups + [
            "hour",
            "minute",
            "second",
        ]

        # ex: 'date__gt'
        for lookup in standard_lookups:
            field, resolved_lookup = resolve_field(
                model_field, LOOKUP_SEP.join(["date", lookup])
            )
            self.assertIsInstance(field, models.DateField)
            self.assertEqual(resolved_lookup, lookup)

        # ex: 'year__iexact'
        for part in datetime_lookups:
            for lookup in standard_lookups:
                field, resolved_lookup = resolve_field(
                    model_field, LOOKUP_SEP.join([part, lookup])
                )
                self.assertIsInstance(field, models.IntegerField)
                self.assertEqual(resolved_lookup, lookup)

        # ex: 'date__year__lte'
        for part in date_lookups:
            for lookup in standard_lookups:
                field, resolved_lookup = resolve_field(
                    model_field, LOOKUP_SEP.join(["date", part, lookup])
                )
                self.assertIsInstance(field, models.IntegerField)
                self.assertEqual(resolved_lookup, lookup)

    def test_resolve_implicit_exact_lookup(self):
        # Use a DateTimeField, so we can check multiple transforms.
        # eg, date__year__gte
        model_field = Article._meta.get_field("published")

        field, lookup = resolve_field(model_field, "date")
        self.assertIsInstance(field, models.DateField)
        self.assertEqual(lookup, "exact")

        field, lookup = resolve_field(model_field, "date__year")
        self.assertIsInstance(field, models.IntegerField)
        self.assertEqual(lookup, "exact")

    def test_invalid_lookup_expression(self):
        model_field = Article._meta.get_field("published")

        with self.assertRaises(FieldLookupError) as context:
            resolve_field(model_field, "invalid_lookup")

        exc = str(context.exception)
        self.assertIn(str(model_field), exc)
        self.assertIn("invalid_lookup", exc)

    def test_invalid_transformed_lookup_expression(self):
        model_field = Article._meta.get_field("published")

        with self.assertRaises(FieldLookupError) as context:
            resolve_field(model_field, "date__invalid_lookup")

        exc = str(context.exception)
        self.assertIn(str(model_field), exc)
        self.assertIn("date__invalid_lookup", exc)


class VerboseFieldNameTests(TestCase):
    def test_none(self):
        verbose_name = verbose_field_name(Article, None)
        self.assertEqual(verbose_name, "[invalid name]")

    def test_invalid_name(self):
        verbose_name = verbose_field_name(Article, "foobar")
        self.assertEqual(verbose_name, "[invalid name]")

    def test_field(self):
        verbose_name = verbose_field_name(Article, "author")
        self.assertEqual(verbose_name, "author")

    def test_field_with_verbose_name(self):
        verbose_name = verbose_field_name(Article, "name")
        self.assertEqual(verbose_name, "title")

    def test_field_all_caps(self):
        verbose_name = verbose_field_name(NetworkSetting, "cidr")
        self.assertEqual(verbose_name, "CIDR")

    def test_forwards_related_field(self):
        verbose_name = verbose_field_name(Article, "author__username")
        self.assertEqual(verbose_name, "author username")

    def test_backwards_related_field(self):
        verbose_name = verbose_field_name(Book, "lovers__first_name")
        self.assertEqual(verbose_name, "lovers first name")

    def test_backwards_related_field_multi_word(self):
        verbose_name = verbose_field_name(User, "manager_of")
        self.assertEqual(verbose_name, "manager of")

    def test_lazy_text(self):
        # sanity check
        field = User._meta.get_field("username")
        self.assertIsInstance(field.verbose_name, Promise)

        verbose_name = verbose_field_name(User, "username")
        self.assertEqual(verbose_name, "username")

    def test_forwards_fk(self):
        verbose_name = verbose_field_name(Article, "author")
        self.assertEqual(verbose_name, "author")

    def test_backwards_fk(self):
        # https://github.com/carltongibson/django-filter/issues/716

        # related_name is set
        verbose_name = verbose_field_name(Company, "locations")
        self.assertEqual(verbose_name, "locations")

        # related_name not set. Auto-generated relation is `article_set`
        # _meta.get_field raises FieldDoesNotExist
        verbose_name = verbose_field_name(User, "article_set")
        self.assertEqual(verbose_name, "[invalid name]")

        # WRONG NAME! Returns ManyToOneRel with related_name == None.
        verbose_name = verbose_field_name(User, "article")
        self.assertEqual(verbose_name, "[invalid name]")


class VerboseLookupExprTests(TestCase):
    def test_exact(self):
        # Exact should default to empty. A verbose expression is unnecessary,
        # and this behavior works well with list syntax for `Meta.fields`.
        verbose_lookup = verbose_lookup_expr("exact")
        self.assertEqual(verbose_lookup, "")

    def test_verbose_expression(self):
        verbose_lookup = verbose_lookup_expr("date__lt")
        self.assertEqual(verbose_lookup, "date is less than")

    def test_missing_keys(self):
        verbose_lookup = verbose_lookup_expr("foo__bar__lt")
        self.assertEqual(verbose_lookup, "foo bar is less than")

    @override_settings(FILTERS_VERBOSE_LOOKUPS={"exact": "is equal to"})
    def test_overridden_settings(self):
        verbose_lookup = verbose_lookup_expr("exact")
        self.assertEqual(verbose_lookup, "is equal to")


class LabelForFilterTests(TestCase):
    def test_standard_label(self):
        label = label_for_filter(Article, "name", "in")
        self.assertEqual(label, "Title is in")

    def test_related_model(self):
        label = label_for_filter(Article, "author__first_name", "in")
        self.assertEqual(label, "Author first name is in")

    def test_exclusion_label(self):
        label = label_for_filter(Article, "name", "in", exclude=True)
        self.assertEqual(label, "Exclude title is in")

    def test_related_model_exclusion(self):
        label = label_for_filter(Article, "author__first_name", "in", exclude=True)
        self.assertEqual(label, "Exclude author first name is in")

    def test_exact_lookup(self):
        label = label_for_filter(Article, "name", "exact")
        self.assertEqual(label, "Title")

    def test_field_all_caps(self):
        label = label_for_filter(NetworkSetting, "cidr", "contains", exclude=True)
        self.assertEqual(label, "Exclude CIDR contains")


@unittest.skipUnless(django.VERSION < (5, 0), "is_dst removed in Django 5.0")
class HandleTimezone(TestCase):
    @override_settings(TIME_ZONE="America/Sao_Paulo")
    def test_handle_dst_ending(self):
        dst_ending_date = datetime.datetime(2017, 2, 18, 23, 59, 59, 999999)
        handled = handle_timezone(dst_ending_date, False)
        self.assertEqual(
            handled,
            make_aware(dst_ending_date, get_default_timezone(), False),
        )

    @override_settings(TIME_ZONE="America/Sao_Paulo")
    def test_handle_dst_starting(self):
        dst_starting_date = datetime.datetime(2017, 10, 15, 0, 0, 0, 0)
        handled = handle_timezone(dst_starting_date, True)
        self.assertEqual(
            handled,
            make_aware(dst_starting_date, get_default_timezone(), True),
        )


class TranslateValidationDataTests(TestCase):
    class F(FilterSet):
        class Meta:
            model = Article
            fields = ["id", "author", "name"]

        choice = MultipleChoiceFilter(choices=[("1", "one"), ("2", "two")])

    def test_error_detail(self):
        f = self.F(
            data={
                "id": "foo",
                "author": "bar",
                "name": "baz",
                "choice": ["3"],
            }
        )
        exc = translate_validation(f.errors)

        self.assertDictEqual(
            exc.detail,
            {
                "id": ["Enter a number."],
                "author": [
                    "Select a valid choice. That choice is not one of the available choices."
                ],
                "choice": [
                    "Select a valid choice. 3 is not one of the available choices."
                ],
            },
        )

    def test_full_error_details(self):
        f = self.F(
            data={
                "id": "foo",
                "author": "bar",
                "name": "baz",
                "choice": ["3"],
            }
        )
        exc = translate_validation(f.errors)

        self.assertEqual(
            exc.get_full_details(),
            {
                "id": [{"message": "Enter a number.", "code": "invalid"}],
                "author": [
                    {
                        "message": "Select a valid choice. That choice is not one of the available choices.",
                        "code": "invalid_choice",
                    }
                ],
                "choice": [
                    {
                        "message": "Select a valid choice. 3 is not one of the available choices.",
                        "code": "invalid_choice",
                    }
                ],
            },
        )



import warnings
from unittest import mock, skipIf

from django.db.models import BooleanField
from django.test import TestCase
from django.test.utils import ignore_warnings, override_settings
from rest_framework import generics, serializers
from rest_framework.test import APIRequestFactory

from django_filters import RemovedInDjangoFilter25Warning, compat, filters
from django_filters.rest_framework import DjangoFilterBackend, FilterSet, backends

from ..models import Article
from .models import CategoryItem, FilterableItem

factory = APIRequestFactory()


class FilterableItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = FilterableItem
        fields = "__all__"


class CategoryItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = CategoryItem
        fields = "__all__"


# These class are used to test a filter class.
class SeveralFieldsFilter(FilterSet):
    text = filters.CharFilter(lookup_expr="icontains")
    decimal = filters.NumberFilter(lookup_expr="lt")
    date = filters.DateFilter(lookup_expr="gt")

    class Meta:
        model = FilterableItem
        fields = ["text", "decimal", "date"]


# Basic filter on a list view.
class FilterableItemView(generics.ListCreateAPIView):
    queryset = FilterableItem.objects.all()
    serializer_class = FilterableItemSerializer
    filter_backends = (DjangoFilterBackend,)


class FilterFieldsRootView(FilterableItemView):
    filterset_fields = ["decimal", "date"]


class FilterClassRootView(FilterableItemView):
    filterset_class = SeveralFieldsFilter


class CategoryItemView(generics.ListCreateAPIView):
    queryset = CategoryItem.objects.all()
    serializer_class = CategoryItemSerializer
    filter_backends = (DjangoFilterBackend,)
    filterset_fields = ["category"]


class GetFilterClassTests(TestCase):
    def test_filterset_class(self):
        class Filter(FilterSet):
            class Meta:
                model = FilterableItem
                fields = "__all__"

        backend = DjangoFilterBackend()
        view = FilterableItemView()
        view.filterset_class = Filter
        queryset = FilterableItem.objects.all()

        filterset_class = backend.get_filterset_class(view, queryset)
        self.assertIs(filterset_class, Filter)

    def test_filterset_class_no_meta(self):
        class Filter(FilterSet):
            pass

        backend = DjangoFilterBackend()
        view = FilterableItemView()
        view.filterset_class = Filter
        queryset = FilterableItem.objects.all()

        filterset_class = backend.get_filterset_class(view, queryset)
        self.assertIs(filterset_class, Filter)

    def test_filterset_class_no_queryset(self):
        class Filter(FilterSet):
            class Meta:
                model = FilterableItem
                fields = "__all__"

        backend = DjangoFilterBackend()
        view = FilterableItemView()
        view.filterset_class = Filter

        filterset_class = backend.get_filterset_class(view, None)
        self.assertIs(filterset_class, Filter)

    def test_filterset_fields(self):
        backend = DjangoFilterBackend()
        view = FilterableItemView()
        view.filterset_fields = ["text", "decimal", "date"]
        queryset = FilterableItem.objects.all()

        filterset_class = backend.get_filterset_class(view, queryset)
        self.assertEqual(filterset_class._meta.fields, view.filterset_fields)

    def test_filterset_fields_malformed(self):
        backend = DjangoFilterBackend()
        view = FilterableItemView()
        view.filterset_fields = ["non_existent"]
        queryset = FilterableItem.objects.all()

        msg = "'Meta.fields' must not contain non-model field names: non_existent"
        with self.assertRaisesMessage(TypeError, msg):
            backend.get_filterset_class(view, queryset)

    def test_filterset_fields_no_queryset(self):
        backend = DjangoFilterBackend()
        view = FilterableItemView()
        view.filterset_fields = ["text", "decimal", "date"]

        filterset_class = backend.get_filterset_class(view, None)
        self.assertIsNone(filterset_class)


@skipIf(compat.coreapi is None, "coreapi must be installed")
class GetSchemaFieldsTests(TestCase):
    def test_fields_with_filterset_fields_list(self):
        backend = DjangoFilterBackend()
        fields = backend.get_schema_fields(FilterFieldsRootView())
        fields = [f.name for f in fields]

        self.assertEqual(fields, ["decimal", "date"])

    def test_filterset_fields_list_with_bad_get_queryset(self):
        """
        See:
          * https://github.com/carltongibson/django-filter/issues/551
        """

        class BadGetQuerySetView(FilterFieldsRootView):
            filterset_fields = ["decimal", "date"]

            def get_queryset(self):
                raise AttributeError("I don't have that")

        backend = DjangoFilterBackend()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            fields = backend.get_schema_fields(BadGetQuerySetView())
            self.assertEqual(
                fields, [], "get_schema_fields should handle AttributeError"
            )

            warning = "{} is not compatible with schema generation".format(
                BadGetQuerySetView
            )
            self.assertEqual(len(w), 1)
            self.assertEqual(str(w[0].message), warning)

    def test_malformed_filterset_fields(self):
        # Malformed filter fields should raise an exception
        class View(FilterFieldsRootView):
            filterset_fields = ["non_existent"]

        backend = DjangoFilterBackend()

        msg = "'Meta.fields' must not contain non-model field names: non_existent"
        with self.assertRaisesMessage(TypeError, msg):
            backend.get_schema_fields(View())

    def test_fields_with_filterset_fields_dict(self):
        class DictFilterFieldsRootView(FilterFieldsRootView):
            filterset_fields = {
                "decimal": ["exact", "lt", "gt"],
            }

        backend = DjangoFilterBackend()
        fields = backend.get_schema_fields(DictFilterFieldsRootView())
        fields = [f.name for f in fields]

        self.assertEqual(fields, ["decimal", "decimal__lt", "decimal__gt"])

    def test_fields_with_filterset_class(self):
        backend = DjangoFilterBackend()
        fields = backend.get_schema_fields(FilterClassRootView())
        schemas = [f.schema for f in fields]
        fields = [f.name for f in fields]

        self.assertEqual(fields, ["text", "decimal", "date"])
        self.assertIsInstance(schemas[0], compat.coreschema.String)
        self.assertIsInstance(schemas[1], compat.coreschema.Number)
        self.assertIsInstance(schemas[2], compat.coreschema.String)

    def test_field_required(self):
        class RequiredFieldsFilter(SeveralFieldsFilter):
            required_text = filters.CharFilter(required=True)

            class Meta(SeveralFieldsFilter.Meta):
                fields = SeveralFieldsFilter.Meta.fields + ["required_text"]

        class FilterClassWithRequiredFieldsView(FilterClassRootView):
            filterset_class = RequiredFieldsFilter

        backend = DjangoFilterBackend()
        fields = backend.get_schema_fields(FilterClassWithRequiredFieldsView())
        required = [f.required for f in fields]
        fields = [f.name for f in fields]

        self.assertEqual(fields, ["text", "decimal", "date", "required_text"])
        self.assertFalse(required[0])
        self.assertFalse(required[1])
        self.assertFalse(required[2])
        self.assertTrue(required[3])

    def tests_field_with_request_callable(self):
        def qs(request):
            # users expect a valid request object to be provided which cannot
            # be guaranteed during schema generation.
            self.fail(
                "callable queryset should not be invoked during schema generation"
            )

        class F(SeveralFieldsFilter):
            f = filters.ModelChoiceFilter(queryset=qs)

        class View(FilterClassRootView):
            filterset_class = F

        view = View()
        view.request = factory.get("/")
        backend = DjangoFilterBackend()
        fields = backend.get_schema_fields(view)
        fields = [f.name for f in fields]

        self.assertEqual(fields, ["text", "decimal", "date", "f"])


class GetSchemaOperationParametersTests(TestCase):
    @ignore_warnings(category=RemovedInDjangoFilter25Warning)
    def test_get_operation_parameters_with_filterset_fields_list(self):
        backend = DjangoFilterBackend()
        fields = backend.get_schema_operation_parameters(FilterFieldsRootView())
        fields = [f["name"] for f in fields]

        self.assertEqual(fields, ["decimal", "date"])

    @ignore_warnings(category=RemovedInDjangoFilter25Warning)
    def test_get_operation_parameters_with_filterset_fields_list_with_choices(self):
        backend = DjangoFilterBackend()
        fields = backend.get_schema_operation_parameters(CategoryItemView())

        self.assertEqual(
            fields,
            [
                {
                    "name": "category",
                    "required": False,
                    "in": "query",
                    "description": "category",
                    "schema": {"type": "string", "enum": ["home", "office"]},
                }
            ],
        )

    def test_deprecation_warning(self):
        backend = DjangoFilterBackend()
        msg = "Built-in schema generation is deprecated. Use drf-spectacular."
        with self.assertWarnsMessage(RemovedInDjangoFilter25Warning, msg):
            backend.get_schema_operation_parameters(FilterFieldsRootView())


class TemplateTests(TestCase):
    def test_backend_output(self):
        """
        Ensure backend renders default if template path does not exist
        """
        view = FilterFieldsRootView()
        backend = view.filter_backends[0]
        request = view.initialize_request(factory.get("/"))
        html = backend().to_html(request, view.get_queryset(), view)

        self.assertHTMLEqual(
            html,
            """
        <h2>Field filters</h2>
        <form class="form" action="" method="get">
            <p>
                <label for="id_decimal">Decimal:</label>
                <input id="id_decimal" name="decimal" step="any" type="number" />
            </p>
            <p>
                <label for="id_date">Date:</label>
                <input id="id_date" name="date" type="text" />
            </p>
            <button type="submit" class="btn btn-primary">Submit</button>
        </form>
        """,
        )

    def test_template_path(self):
        view = FilterFieldsRootView()

        class Backend(view.filter_backends[0]):
            template = "filter_template.html"

        request = view.initialize_request(factory.get("/"))
        html = Backend().to_html(request, view.get_queryset(), view)

        self.assertHTMLEqual(html, "Test")

    @override_settings(TEMPLATES=[])
    def test_DTL_missing(self):
        # The backend should be importable even if the DTL is not used.
        # See: https://github.com/carltongibson/django-filter/issues/506
        try:
            from importlib import reload  # python 3.4
        except ImportError:
            from imp import reload

        reload(backends)

    def test_multiple_engines(self):
        # See: https://github.com/carltongibson/django-filter/issues/578
        DTL = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "APP_DIRS": True,
        }
        ALT = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "APP_DIRS": True,
            "NAME": "alt",
        }

        # multiple DTL backends
        with override_settings(TEMPLATES=[DTL, ALT]):
            self.test_backend_output()


class AutoFilterSetTests(TestCase):
    def test_autofilter_meta_inheritance(self):
        # https://github.com/carltongibson/django-filter/issues/663

        class F(FilterSet):
            class Meta:
                filter_overrides = {BooleanField: {}}

        class Backend(DjangoFilterBackend):
            filterset_base = F

        view = FilterFieldsRootView()
        backend = Backend()

        filterset_class = backend.get_filterset_class(view, view.get_queryset())
        filter_overrides = filterset_class._meta.filter_overrides

        # derived filterset_class.Meta should inherit from default_filter_set.Meta
        self.assertIn(BooleanField, filter_overrides)
        self.assertDictEqual(filter_overrides[BooleanField], {})


class ValidationErrorTests(TestCase):
    def test_errors(self):
        class F(FilterSet):
            class Meta:
                model = Article
                fields = ["id", "author", "name"]

        view = FilterFieldsRootView()
        backend = DjangoFilterBackend()
        request = factory.get("/?id=foo&author=bar&name=baz")
        request = view.initialize_request(request)
        queryset = Article.objects.all()
        view.filterset_class = F

        with self.assertRaises(serializers.ValidationError) as exc:
            backend.filter_queryset(request, queryset, view)

        # test output, does not include error code
        self.assertDictEqual(
            exc.exception.detail,
            {
                "id": ["Enter a number."],
                "author": [
                    "Select a valid choice. That choice is not one of the available choices."
                ],
            },
        )


class DjangoFilterBackendTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.backend = DjangoFilterBackend()
        cls.backend.get_filterset_class = lambda x, y: None

    def test_get_filterset_none_filter_class(self):
        filterset = self.backend.get_filterset(mock.Mock(), mock.Mock(), mock.Mock())
        self.assertIsNone(filterset)

    def test_filter_queryset_none_filter_class(self):
        prev_qs = mock.Mock()
        qs = self.backend.filter_queryset(mock.Mock(), prev_qs, mock.Mock())
        self.assertIs(qs, prev_qs)

    def test_to_html_none_filter_class(self):
        html = self.backend.to_html(mock.Mock(), mock.Mock(), mock.Mock())
        self.assertIsNone(html)

    @ignore_warnings(category=RemovedInDjangoFilter25Warning)
    def test_get_schema_operation_parameters_userwarning(self):
        with self.assertWarns(UserWarning):
            view = mock.Mock()
            view.__class__.return_value = "Test"
            view.get_queryset.side_effect = Exception
            self.backend.get_schema_operation_parameters(view)

    @mock.patch("django_filters.compat.is_crispy", return_value=True)
    def test_template_crispy(self, _):
        self.assertEqual(
            self.backend.template, "django_filters/rest_framework/crispy_form.html"
        )



import warnings

from django.template import loader

from .. import compat, utils
from . import filters, filterset


class DjangoFilterBackend:
    filterset_base = filterset.FilterSet
    raise_exception = True

    @property
    def template(self):
        if compat.is_crispy():
            return "django_filters/rest_framework/crispy_form.html"
        return "django_filters/rest_framework/form.html"

    def get_filterset(self, request, queryset, view):
        filterset_class = self.get_filterset_class(view, queryset)
        if filterset_class is None:
            return None

        kwargs = self.get_filterset_kwargs(request, queryset, view)
        return filterset_class(**kwargs)

    def get_filterset_class(self, view, queryset=None):
        """
        Return the `FilterSet` class used to filter the queryset.
        """
        filterset_class = getattr(view, "filterset_class", None)
        filterset_fields = getattr(view, "filterset_fields", None)

        if filterset_class:
            filterset_model = filterset_class._meta.model

            # FilterSets do not need to specify a Meta class
            if filterset_model and queryset is not None:
                assert issubclass(
                    queryset.model, filterset_model
                ), "FilterSet model %s does not match queryset model %s" % (
                    filterset_model,
                    queryset.model,
                )

            return filterset_class

        if filterset_fields and queryset is not None:
            MetaBase = getattr(self.filterset_base, "Meta", object)

            class AutoFilterSet(self.filterset_base):
                class Meta(MetaBase):
                    model = queryset.model
                    fields = filterset_fields

            return AutoFilterSet

        return None

    def get_filterset_kwargs(self, request, queryset, view):
        return {
            "data": request.query_params,
            "queryset": queryset,
            "request": request,
        }

    def filter_queryset(self, request, queryset, view):
        filterset = self.get_filterset(request, queryset, view)
        if filterset is None:
            return queryset

        if not filterset.is_valid() and self.raise_exception:
            raise utils.translate_validation(filterset.errors)
        return filterset.qs

    def to_html(self, request, queryset, view):
        filterset = self.get_filterset(request, queryset, view)
        if filterset is None:
            return None

        template = loader.get_template(self.template)
        context = {"filter": filterset}
        return template.render(context, request)

    def get_coreschema_field(self, field):
        if isinstance(field, filters.NumberFilter):
            field_cls = compat.coreschema.Number
        else:
            field_cls = compat.coreschema.String
        return field_cls(description=str(field.extra.get("help_text", "")))

    def get_schema_fields(self, view):
        # This is not compatible with widgets where the query param differs from the
        # filter's attribute name. Notably, this includes `MultiWidget`, where query
        # params will be of the format `<name>_0`, `<name>_1`, etc...
        from django_filters import RemovedInDjangoFilter25Warning
        warnings.warn(
            "Built-in schema generation is deprecated. Use drf-spectacular.",
            category=RemovedInDjangoFilter25Warning,
        )
        assert (
            compat.coreapi is not None
        ), "coreapi must be installed to use `get_schema_fields()`"
        assert (
            compat.coreschema is not None
        ), "coreschema must be installed to use `get_schema_fields()`"

        try:
            queryset = view.get_queryset()
        except Exception:
            queryset = None
            warnings.warn(
                "{} is not compatible with schema generation".format(view.__class__)
            )

        filterset_class = self.get_filterset_class(view, queryset)

        return (
            []
            if not filterset_class
            else [
                compat.coreapi.Field(
                    name=field_name,
                    required=field.extra["required"],
                    location="query",
                    schema=self.get_coreschema_field(field),
                )
                for field_name, field in filterset_class.base_filters.items()
            ]
        )

    def get_schema_operation_parameters(self, view):
        from django_filters import RemovedInDjangoFilter25Warning
        warnings.warn(
            "Built-in schema generation is deprecated. Use drf-spectacular.",
            category=RemovedInDjangoFilter25Warning,
        )
        try:
            queryset = view.get_queryset()
        except Exception:
            queryset = None
            warnings.warn(
                "{} is not compatible with schema generation".format(view.__class__)
            )

        filterset_class = self.get_filterset_class(view, queryset)

        if not filterset_class:
            return []

        parameters = []
        for field_name, field in filterset_class.base_filters.items():
            parameter = {
                "name": field_name,
                "required": field.extra["required"],
                "in": "query",
                "description": field.label if field.label is not None else field_name,
                "schema": {
                    "type": "string",
                },
            }
            if field.extra and "choices" in field.extra:
                parameter["schema"]["enum"] = [c[0] for c in field.extra["choices"]]
            parameters.append(parameter)
        return parameters
